from bspump.asab import Service, Config
import concurrent.futures


class ProactorService(Service):
    """
    Proactor service is useful for running CPU bound operations from asynchronous part of the code that would potentially block the main thread.
    It allows to run these processes from different threads.
    """

    def __init__(self, app, service_name):
        super().__init__(app, service_name)
        self.Loop = app.Loop

        max_workers = Config.get("asab:proactor", "max_workers")
        if max_workers:
            max_workers = int(max_workers)
        else:
            max_workers = None
        if max_workers <= 0:
            max_workers = None

        self.Executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=max_workers,  # The maximum number of threads that can be used to execute the given calls.
            # If None, ThreadPoolExecutor will determine the number itself based on number of CPU's.
            thread_name_prefix="AsabProactorThread",
        )

        if Config.get("asab:proactor", "default_executor"):
            self.Loop.set_default_executor(self.Executor)

    # There was the method run, which is obsolete
    def execute(self, func, *args):
        """
        Execute `func(*args)` in the thread from the Proactor Service pool.
        Return Future or Task that must be awaited and it provides the result of the `func()` call.
        """
        return self.Loop.run_in_executor(self.Executor, func, *args)

    def schedule(self, func, *args):
        """
        Execute `func(*args)` in the thread from the Proactor Service pool.
        The result of the future is discarded (using Task Service).
        """

        future = self.execute(func, *args)
        self.App.TaskService.schedule(future)



from .trigger import Trigger


class PubSubTrigger(Trigger):
    def __init__(self, app, message_types, pubsub=None, id=None):
        super().__init__(app, id)
        self.PubSub = pubsub if pubsub is not None else app.PubSub

        if isinstance(message_types, str):
            self.PubSub.subscribe(message_types, self.on_message)
        else:
            for message_type in message_types:
                self.PubSub.subscribe(message_type, self.on_message)

    async def on_message(self, message_type):
        self.fire()

    @classmethod
    def construct(cls, app, definition: dict):
        id = definition.get("id")
        message_types = definition.get("args", {}).get("message_types")
        return cls(app, id=id, message_types=message_types)



import os
import datetime

import bspump.declarative
import bspump.unittest


class TestDeclarativeDateTime(bspump.unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Builder = bspump.declarative.ExpressionBuilder(self.App)

    def load(self, decl_fname):
        basedir = os.path.dirname(__file__)
        with open(os.path.join(basedir, decl_fname), "r") as f:
            return self.Builder.parse(f.read())[0]

    def test_datetime_parse_set_year(self):
        decl = self.load("./test_datetime_parse_set_year.yaml")

        res = decl({}, {}, "Mar 16 07:46:24")
        now = datetime.datetime.now()
        dt = datetime.datetime(
            year=now.year,
            month=3,
            day=16,
            hour=7,
            minute=46,
            second=24,
            tzinfo=datetime.timezone.utc,
        )
        self.assertEqual(res, dt.timestamp())

    def test_datetime_parse_timezone(self):
        decl = self.load("./test_datetime_parse_timezone.yaml")

        res = decl({}, {})
        self.assertEqual(res, 946681200.0)

    def test_datetime_parse_add(self):
        decl = self.load("./test_datetime_add.yaml")

        res = decl({}, {})
        self.assertEqual(res, 2.0)

    def test_datetime_parse_failed(self):
        decl = self.load("./test_datetime_parse_fail.yaml")

        res = decl({}, {})
        self.assertEqual(res, "Failed :-(")

    def test_datetime_parse_tz(self):
        decl = self.load("./test_datetime_parse_tz.yaml")

        res = decl({}, {})
        self.assertEqual(res, 1585621784.0)



import logging

from ..abc.service import Service

#

L = logging.getLogger(__name__)

#


class ZooKeeperService(Service):
    def __init__(self, app, service_name):
        super().__init__(app, service_name)

        # Make sure that the proactor service exists
        from ..proactor import Module

        app.add_module(Module)
        self.ProactorService = app.get_service("asab.ProactorService")

        self.Containers = []

    async def finalize(self, app):
        # Remove containers from the list
        while len(self.Containers) > 0:
            container = self.Containers.pop()
            await container._stop()



from ..abc.processor import Processor
import random


class RandomDrop(Processor):
    """
    Simulates random drop of events.
    """

    ConfigDefaults = {
        "threshold": 0.5,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Threshold = float(self.Config["threshold"])

    def process(self, context, event):
        if random.random() < self.Threshold:
            return None
        return event



from .sink import AvroSink
from .source import AvroSource
from .serializer import AvroSerializer
from .deserializer import AvroDeserializer

__all__ = (
    "AvroSink",
    "AvroSource",
    "AvroSerializer",
    "AvroDeserializer",
)



from ...abc import Expression, evaluate


class LIST(Expression):
    Attributes = {
        "With": ["*"],  # TODO: This ...
        "Append": ["*"],  # TODO: This ...
    }

    def __init__(self, app, *, arg_with=None, arg_append=None):
        super().__init__(app)

        self.With = arg_with

        if arg_append is not None:
            assert isinstance(arg_append, list)
        self.Append = arg_append

    def __call__(self, context, event, *args, **kwargs):
        if self.With is None:
            with_list = list()
        else:
            with_list = evaluate(self.With, context, event, *args, **kwargs)
            if not isinstance(with_list, list):
                with_list = list()

        if self.Append is not None:
            for value in self.Append:
                with_list.append(evaluate(value, context, event, *args, **kwargs))

        return with_list



import logging
import json
import socket
import typing

import aiohttp

from .. import Service


L = logging.getLogger(__name__)


class DiscoveryService(Service):
    def __init__(self, app, zkc, service_name="asab.DiscoveryService") -> None:
        super().__init__(app, service_name)
        self.ZooKeeperContainer = zkc

    async def locate(self, instance_id: str = None, service_id: str = None) -> list:
        """
        Returns a list of URLs for a given instance or service ID.

        :param instance_id: The ID of a specific instance of a service that the client wants to locate.
        :type instance_id: str
        :param service_id: The `service_id` parameter represents identifier of a
        service to locate. It is used to query a service registry to find the
        instances of the service that are currently available.
        :type service_id: str
        :return: A list of URLs in the format "http://servername:port" for the specified instance or
        service.
        """
        return [
            "http://{}:{}".format(servername, port)
            for servername, port in await self._locate(instance_id, service_id)
        ]

    async def _locate(
        self, instance_id: str = None, service_id: str = None
    ) -> typing.List[typing.Tuple]:
        """
        Locates service instances based on their instance ID or service ID.

        :param instance_id: The unique identifier for a specific instance of a service
        :type instance_id: str
        :param service_id: The ID of the service to locate
        :type service_id: str
        :return: a list of tuples containing the server name and port number of the located service(s).
        """
        if instance_id is None and service_id is None:
            L.warning(
                "Please provide instance_id, service_id, or appclass to locate the service(s)."
            )
            return

        instances = await self.get_advertised_instances()
        if len(instances) == 0:
            L.warning("No instances available.")
            return
        res = []
        for instance in instances:
            if service_id is not None:
                if service_id != instance.get("service_id", "").lower():
                    continue

            if instance_id is not None:
                if instance_id != instance.get("instance_id", "").lower():
                    continue

            web = instance.get("web")
            host = instance.get("node_id", instance.get("host"))

            if web is None or host is None:
                continue
            for i in web:
                try:
                    ip = i[0]
                    port = i[1]
                except KeyError:
                    L.error(
                        "Unexpected format of 'web' section in advertised data: '{}'".format(
                            web
                        )
                    )
                    return
                if ip not in ("0.0.0.0", "::"):
                    continue
                res.append((host, port))
        return res

    async def get_advertised_instances(self) -> typing.List[typing.Dict]:
        """
        Returns a list of dictionaries. Each dictionary represents an advertised instance
        obtained by iterating over the items in the `/run` path in ZooKeeper.
        """
        advertised = []
        async for item, item_data in self._iter_zk_items("/run"):
            data_dict = json.loads(item_data)
            data_dict["zookeeper_id"] = item
            advertised.append(data_dict)

        return advertised

    async def _iter_zk_items(self, path):
        base_path = self.ZooKeeperContainer.Path + path
        items = await self.ZooKeeperContainer.ZooKeeper.get_children(base_path)
        if items is None:
            L.error(
                "Missing '{}/run' folder in ZooKeeper.".format(
                    self.ZooKeeperContainer.Path
                )
            )
            return

        for item in items:
            item_data = await self.ZooKeeperContainer.ZooKeeper.get_data(
                base_path + "/" + item
            )
            yield item, item_data

    def session(self):
        """
        Usage:

        async with self.DiscoveryService.session() as session:
                # use URL in format: <protocol>://<value>.<key>.asab/<endpoint> where key is "service_id" or "instance_id" and value the respective serivce identificator
                async with session.get("http://my_application_1.instance_id.asab/asab/v1/config") as resp:
                        ...
        """
        return aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(resolver=DiscoveryResolver(self))
        )


class DiscoveryResolver(aiohttp.DefaultResolver):
    """Custom aiohttp Resolver for Discovery Session based on default aiohttp resolver."""

    def __init__(self, svc) -> None:
        super().__init__()
        self.DiscoveryService = svc

    async def resolve(
        self, hostname: str, port: int = 0, family: int = socket.AF_INET
    ) -> typing.List[typing.Dict[str, typing.Any]]:
        """
        Resolves a hostname only with '.asab' domain. and returns a list of dictionaries
        containing information about the resolved hosts further used by aiohttp.TCPConnector

        The hostname to resolve must be in the format of "<value>.<key>.asab",
        where key is "service_id" or "instance_id" and value is the particular identificator of the service to be resolved.
        The "asab" domain is required for resolution, otherwise it is treated like normal URL.
        """
        url_split = hostname.rsplit(".", 2)

        # If there is no asab domain, it is normal URL and resolve it normally
        if url_split[-1] != "asab":
            return await super().resolve(hostname, port, family)

        # Make sure the format of the hostname is right
        if len(url_split) != 3:
            raise NotDiscoveredError(
                "Invalid format of the hostname '{}'. Use e.g. `asab-config.service_id.asab` instead.".format(
                    hostname
                )
            )

        hosts = []
        located_instances = await self.DiscoveryService._locate(
            **{url_split[1]: url_split[0]}
        )
        if located_instances is None or len(located_instances) == 0:
            raise NotDiscoveredError("Failed to discover '{}'.".format(hostname))
        for i in located_instances:
            hosts.append(*await super().resolve(i[0], i[1], family))

        return hosts


class NotDiscoveredError(RuntimeError):
    pass



import abc
import logging
import os
import numpy as np

from bspump.asab import Configurable
from .utils.closedrows import ClosedRows, PersistentClosedRows

###

L = logging.getLogger(__name__)

###


class Matrix(abc.ABC, Configurable):
    """
    Generic `Matrix` object.

    Matrix structure is organized in a following hiearchical order:

    Matrix -> Rows -> Columns -> Cells

    Cells have unified data format across the whole matrix.
    This format is specified by a `dtype`.
    It can be a simple integer or float but also a complex dictionary type with names and types of the fields.

    The description of types that can be used for a `dtype` of a cell:

    +------------+------------------+
    | Name       | Definition       |
    +============+==================+
    | 'b'        | Byte             |
    +------------+------------------+
    | 'i'        | Signed integer   |
    +------------+------------------+
    | 'u'        | Unsigned integer |
    +------------+------------------+
    | 'f'        | Floating point   |
    +------------+------------------+
    | 'c'        | Complex floating |
    |            | point            |
    +------------+------------------+
    | 'S'        | String           |
    +------------+------------------+
    | 'U'        | Unicode string   |
    +------------+------------------+
    | 'V'        | Raw data         |
    +------------+------------------+

    Example: 'i8' stands for int64.

    For more details, see https://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html


    Object main attributes:
    `Array` is numpy ndarray, the actual data representation of the matrix object.
    `ClosedRows` is a set, where some row ids can be stored before deletion during the matrix rebuild.

    """

    ConfigDefaults = {
        "max_closed_rows_capacity": 0.2,
    }

    def __init__(self, app, dtype="float_", persistent=False, id=None, config=None):
        if not isinstance(dtype, str):
            dtype = dtype[:]

        self.Id = id if id is not None else self.__class__.__name__
        super().__init__("matrix:{}".format(self.Id), config=config)

        self.App = app
        self.Loop = app.Loop

        self.DType = dtype
        self.MaxClosedRowsCapacity = float(self.Config["max_closed_rows_capacity"])
        self.zeros()

        metrics_service = app.get_service("asab.MetricsService")
        self.Gauge = metrics_service.create_gauge(
            "RowCounter",
            tags={
                "matrix": self.Id,
            },
            init_values={
                "rows.closed": 0,
                "rows.active": 0,
            },
        )

    def zeros(self, rows=1):
        self.Array = np.zeros(self.build_shape(rows), dtype=self.DType)
        self.ClosedRows = ClosedRows()
        self.ClosedRows.add(0)

    def flush(self):
        """
        The matrix will be recreated without rows from `ClosedRows`.
        """
        indexes = set(range(self.Array.shape[0]))
        closed_indexes = set()
        closed_indexes |= self.ClosedRows.get_rows()
        saved_indexes = list(indexes - closed_indexes)
        saved_indexes.sort()
        self.Array = self.Array.take(saved_indexes, axis=0)
        self.ClosedRows.flush(self.Array.shape[0])
        self.Gauge.set("rows.closed", 0)
        self.Gauge.set("rows.active", self.Array.shape[0])
        return closed_indexes, saved_indexes

    def close_rows(self, row_names, clear=True):
        pass

    def close_row(self, row_name, clear=True):
        pass

    def add_row(self):
        try:
            return self.ClosedRows.pop()
        except KeyError:
            self._grow_rows(max(5, int(0.10 * self.Array.shape[0])))
            return self.ClosedRows.pop()
        finally:
            crc = len(self.ClosedRows)
            self.Gauge.set("rows.active", self.Array.shape[0] - crc)
            self.Gauge.set("rows.closed", crc)

    def build_shape(self, rows=0):
        """
        Override this method to have a control over the shape of the matrix.
        """
        return (rows,)

    def reshape(self, shape):
        return shape

    def _grow_rows(self, rows=1):
        """
        Override this method to gain control on how a new closed rows are added to the matrix
        """
        current_rows = self.Array.shape[0]
        self.Array = np.pad(
            self.Array.copy(), ((0, rows), (0, 0)), "constant", constant_values=np.nan
        )
        self.ClosedRows.extend(current_rows, self.Array.shape[0])

    def time(self):
        return self.App.time()

    async def analyze(self):
        """
        The `Matrix` itself can run the `analyze()`.
        It is not recommended to iterate through the matrix row by row (or cell by cell).
        Instead use numpy fuctions. Examples:
        1. You have a vector with n rows. You need only those row indeces, where the cell content is more than 10.
        Use `np.where(vector > 10)`.
        2. You have a matrix with n rows and m columns. You need to find out which rows
        fully consist of zeros. use `np.where(np.all(matrix == 0, axis=1))` to get those row indexes.
        Instead `np.all()` you can use `np.any()` to get all row indexes, where there is at least one zero.
        3. Use `np.mean(matrix, axis=1)` to get means for all rows.
        4. Usefull numpy functions: `np.unique()`, `np.sum()`, `np.argmin()`, `np.argmax()`.
        """
        pass


class PersistentMatrix(Matrix):
    ConfigDefaults = {
        "path": "",
    }

    def __init__(self, app, dtype="float_", id=None, config=None):
        super().__init__(app, dtype=dtype, id=id, config=config)

    def create_path(self):
        self.Path = self.Config["path"]
        if not os.path.exists(self.Path):
            os.makedirs(self.Path)

        self.ArrayPath = os.path.join(self.Path, "array.dat")

    def zeros(self, rows=1):
        self.create_path()
        if os.path.exists(self.ArrayPath):
            self.Array = np.memmap(self.ArrayPath, dtype=self.DType, mode="readwrite")
            self.Array = self.Array.reshape(self.reshape(self.Array.shape))
        else:
            array = np.zeros(self.build_shape(rows), dtype=self.DType)
            self.Array = np.memmap(
                self.ArrayPath, dtype=self.DType, mode="w+", shape=array.shape
            )

        path = os.path.join(self.Path, "closed_rows.dat")
        self.ClosedRows = PersistentClosedRows(path, size=self.Array.shape[0])

    def flush(self):
        """
        The matrix will be recreated without rows from `ClosedRows`.
        """
        indexes = set(range(self.Array.shape[0]))
        closed_indexes = set()
        closed_indexes |= self.ClosedRows.get_rows()
        saved_indexes = list(indexes - closed_indexes)
        saved_indexes.sort()
        self.Array = self.Array.take(saved_indexes, axis=0)
        array = np.memmap(
            self.ArrayPath, dtype=self.DType, mode="w+", shape=self.Array.shape
        )
        array[:] = self.Array[:]
        self.Array = array

        self.ClosedRows.flush(self.Array.shape[0])
        self.Gauge.set("rows.closed", 0)
        self.Gauge.set("rows.active", self.Array.shape[0])
        return closed_indexes, saved_indexes

    def close_row(self, row_name, clear=True):
        row_index = self.Index.get_row_index(row_name)
        if (row_index in self.ClosedRows) or (row_index is None):
            return False

        if clear:
            self.Array[row_index] = np.zeros(1, dtype=self.DType)

        self.ClosedRows.add(row_index)

        if len(self.ClosedRows) >= self.MaxClosedRowsCapacity * self.Array.shape[0]:
            self.flush()

        crc = len(self.ClosedRows)
        self.Gauge.set("rows.active", self.Array.shape[0] - crc)
        self.Gauge.set("rows.closed", crc)
        return True

    def _grow_rows(self, rows=1):
        """
        Override this method to gain control on how a new closed rows are added to the matrix
        """
        current_rows = self.Array.shape[0]
        array = np.zeros(self.Array.shape, dtype=self.DType)
        array[:] = self.Array[:]
        array.resize((current_rows + rows,) + self.Array.shape[1:], refcheck=False)
        self.Array = np.memmap(
            self.ArrayPath, dtype=self.DType, mode="w+", shape=array.shape
        )
        self.Array[:] = array[:]
        self.ClosedRows.extend(current_rows, self.Array.shape[0])



import bspump.unittest
import bspump.common


class TestHexlifyProcessor(bspump.unittest.ProcessorTestCase):
    def test_hexlify_processor(self):
        events = {
            (None, b"\x124Vx"),
            (None, b"\xab\xba\xba\xbe"),
            (None, b"\xde\xad\xc0\xde"),
        }

        self.set_up_processor(bspump.common.HexlifyProcessor)

        output = self.execute(events)

        self.assertEqual(
            sorted([event for context, event in output]),
            [b"12345678", b"abbababe", b"deadc0de"],
        )



import functools

import aiohttp.web

from bspump.asab.web.rest import json_response


def tenant_middleware_factory(app, svc):
    """
    Ensures the tenant is obtained from the TenantService.
    :param app: application object
    :param svc: TenantService
    :return: handler(request)
    """

    @aiohttp.web.middleware
    async def tenant_middleware(request, handler):
        tenant_id = request.match_info.get("tenant")
        if tenant_id is not None:
            tenant = svc.locate_tenant(tenant_id)
            if tenant is not None:
                request.Tenant = tenant
        return await handler(request)

    return tenant_middleware


def tenant_handler(func):
    @functools.wraps(func)
    async def wrapper(*args, **kargs):
        request = args[-1]
        try:
            kargs["tenant"] = request.Tenant
        except AttributeError:
            return json_response(
                request,
                {
                    "result": "TENANT-NOT-FOUND",
                },
                status=404,
            )
        return await func(*args, **kargs)

    return wrapper



from .connection import AMQPConnection
from .source import AMQPSource, AMQPFullMessageSource
from .sink import AMQPSink

__all__ = (
    "AMQPConnection",
    "AMQPSource",
    "AMQPFullMessageSource",
    "AMQPSink",
)



from aiohttp.web import Request


class BaseField:
    def __init__(self, name, **kwargs):
        self.name = name
        self.hidden: bool = kwargs.get("hidden", False)
        self.required: bool = kwargs.get("required", False if self.hidden else True)
        self.display: str = kwargs.get("display", self.name)
        self.description: str = kwargs.get("description", "")
        self.default = kwargs.get("default", "")

    def html(self, defaults) -> str:
        pass

    def get_params(self) -> dict:
        pass

    def restructure_data(self, dfrom, dto):
        pass

    def clean(self, data, request: Request = None):
        pass



from aiohttp.web import Request
from bspump.http.web.components.field import Field
from bspump.http.web.template_env import env


class DateField(Field):
    """
    Renders an HTML5 <input type="date"> and
    cleans the incoming string into the same format.
    """

    def inner_html(self, default="", readonly=False):
        # default should be an ISO date string "YYYY-MM-DD"
        template = env.get_template("date-field.html")
        return template.render(
            default=default,
            default_classes=self.default_classes,
            default_input_props=self.default_input_props,
            readonly=readonly,
        )

    def clean(self, data, request: Request | None = None):
        # leave as string; browser enforces format
        value = data.get(self.name)
        if value is not None:
            # optionally, you could validate format here
            data[self.name] = value



from ...abc import Expression

from ..value.valueexpr import VALUE


class IF(Expression):
    """
    Checks "if" condition passes - if so, proceeds with "then" expression, otherwise with "else"
    """

    Attributes = {
        "Test": ["bool"],  # Test expression MUST return boolean
        "Then": None,  # Determined in initialize()
        "Else": None,  # Determined in initialize()
    }

    Category = "Statements"

    def __init__(self, app, *, arg_test, arg_then=True, arg_else=False):
        super().__init__(app)

        if isinstance(arg_test, Expression):
            self.Test = arg_test
        else:
            self.Test = VALUE(app, value=arg_test)

        if isinstance(arg_then, Expression):
            self.Then = arg_then
        else:
            self.Then = VALUE(app, value=arg_then)

        if isinstance(arg_else, Expression):
            self.Else = arg_else
        else:
            self.Else = VALUE(app, value=arg_else)

        self.Attributes = IF.Attributes.copy()

    def initialize(self):
        # Then and else expression must return the same type
        assert self.Then.get_outlet_type() == self.Else.get_outlet_type()
        self.Attributes["Then"] = self.Then.get_outlet_type()
        self.Attributes["Else"] = self.Else.get_outlet_type()

    def __call__(self, context, event, *args, **kwargs):
        if self.Test(context, event, *args, **kwargs):
            return self.Then(context, event, *args, **kwargs)
        else:
            return self.Else(context, event, *args, **kwargs)

    def get_outlet_type(self):
        return self.Then.get_outlet_type()



from .test_analyzer import *
from .test_geoanalyzer import *
from .test_latch import *
from .test_timedriftanalyzer import *
from .test_timewindowanalyzer import *
from .test_sessionanalyzer import *



# Arithmetic
from .arithmetic import ADD
from .arithmetic import DIV
from .arithmetic import MUL
from .arithmetic import SUB
from .arithmetic import MOD
from .arithmetic import POW

# Logical
from .logical import AND
from .logical import OR
from .logical import NOT

# Comparison
from .comparison import LT
from .comparison import LE
from .comparison import EQ
from .comparison import NE
from .comparison import GE
from .comparison import GT
from .comparison import IS
from .comparison import ISNOT

# Statements
from .statement.ifexpr import IF
from .statement.whenexpr import WHEN
from .statement.forexpr import FOR
from .statement.firstexpr import FIRST
from .statement.selfexpr import SELF
from .statement.funexpr import FUNCTION

# String
from .string.endswith import ENDSWITH
from .string.joinexpr import JOIN
from .string.regex import REGEX
from .string.regex import REGEX_PARSE
from .string.regex import REGEX_REPLACE
from .string.regex import REGEX_SPLIT
from .string.regex import REGEX_FINDALL
from .string.startswith import STARTSWITH
from .string.upperexpr import UPPER
from .string.lowerexpr import LOWER
from .string.substringexpr import SUBSTRING
from .string.cutexpr import CUT
from .string.contains import CONTAINS
from .string.split import SPLIT

# Data structures
from .datastructs.dictexpr import DICT
from .datastructs.dict_parse import DICT_PARSE
from .datastructs.tupleexpr import TUPLE
from .datastructs.listexpr import LIST

from .datastructs.itemexpr import ITEM

# Value
from .value.valueexpr import VALUE
from .value.eventexpr import EVENT
from .value.eventexpr import KWARGS
from .value.eventexpr import KWARG
from .value.eventexpr import ARGS
from .value.eventexpr import ARG

from .utility.context import CONTEXT
from .utility.context import CONTEXT_SET

from .utility.hashexpr import HASH

# Lookup
from .lookup.lookupexpr import LOOKUP_GET
from .lookup.lookupexpr import LOOKUP_CONTAINS

# Test
from .test.inexpr import IN

# IP
from .ip.ipparseexpr import IP_PARSE
from .ip.ipformatexpr import IP_FORMAT
from .ip.insubnetexpr import IP_INSUBNET

# Utility
from .utility.castexpr import CAST
from .utility.mapexpr import MAP
from .utility.debugexpr import DEBUG

# Date/time
from .datetime.nowexpr import NOW
from .datetime.dtformat import DATETIME_FORMAT
from .datetime.dtparse import DATETIME_PARSE
from .datetime.dtget import DATETIME_GET

__all__ = [
    "ADD",
    "DIV",
    "MUL",
    "SUB",
    "MOD",
    "POW",  # Aritmetics
    "AND",
    "OR",
    "NOT",  # Logical
    "LT",
    "LE",
    "EQ",
    "NE",
    "GE",
    "GT",
    "IS",
    "ISNOT",  # Comparison
    "IF",
    "WHEN",
    "FOR",
    "FIRST",
    "SELF",
    "FUNCTION",
    "DEBUG",
    "DICT",
    "LIST",
    "TUPLE",
    "ITEM",
    "DICT_PARSE",
    "STARTSWITH",
    "ENDSWITH",
    "SUBSTRING",
    "UPPER",
    "LOWER",
    "JOIN",
    "CUT",
    "CONTAINS",
    "REGEX",
    "REGEX_PARSE",
    "REGEX_REPLACE",
    "REGEX_SPLIT",
    "REGEX_FINDALL",
    "SPLIT",
    "VALUE",
    "EVENT",
    "CONTEXT",
    "CONTEXT_SET",
    "HASH",
    "ARGS",
    "ARG",
    "KWARGS",
    "KWARG",
    "LOOKUP_GET",
    "LOOKUP_CONTAINS",
    "IN",
    "IP_PARSE",
    "IP_FORMAT",
    "IP_INSUBNET",
    "CAST",
    "MAP",
    "DATETIME_FORMAT",
    "DATETIME_PARSE",
    "DATETIME_GET",
    "NOW",
]



import logging
import aiohttp.web

from bspump.asab import Service

from .session import Session

#

L = logging.getLogger(__name__)

#

# TODO: Configure cookie name
# TODO: Configure cookie http only / secure
# TODO: Configure cookie domain
# TODO: Configure max-age


class ServiceWebSession(Service):
    def __init__(
        self, app, service_name, webapp, session_storage=None, session_class=None
    ):
        super().__init__(app, service_name)

        if session_class is None:
            session_class = Session

        # Construct session storage
        if session_storage is None:
            from .inmemstor import InMemorySessionStorage

            self.SessionStorage = InMemorySessionStorage(app, session_class)
        else:
            self.SessionStorage = session_storage

        # Add middleware to a webservice
        webapp.middlewares.append(session_middleware(self.SessionStorage))


def session_middleware(storage):
    @aiohttp.web.middleware
    async def factory(request, handler):
        session = await storage.load(request)
        request["Session"] = session
        try:
            response = await handler(request)
            if session.Id is None:
                await storage.set(session, response)
            return response
        finally:
            await storage.store(session)

    return factory



import bspump
import collections.abc


class KafkaKeyFilter(bspump.Processor):
    """
    KafkaKeyFilter reduces the incoming event stream from Kafka based on a
    key provided in each event.

    Every Kafka message has a key, KafkaKeyFilter selects only those events where
    the event key matches one of provided 'keys', other events will be discarded.

    Set filtering `keys` as a parameter (in bytes) in the KafkaKeyFilter constructor.

    KafkaKeyFilter is meant to be inserted after KafkaSource in a Pipeline.
    """

    def __init__(self, app, pipeline, keys, id=None, config=None):
        """
        Initializes variables

        **Parameters**

        app : Application
                Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html`_.

        pipeline : Pipeline
                Name of the Pipeline.

        keys : bytes
                keys used to filter out events from the event stream.

        id : , default = None

        config : JSON, default = None
                configuration file in JSON

        """
        super().__init__(app, pipeline, id, config)
        if not isinstance(keys, collections.abc.Iterable) or isinstance(keys, bytes):
            self.Keys = frozenset([keys])
        else:
            self.Keys = frozenset(keys)

    def process(self, context, event):
        kafka_key = context.get("kafka_key")
        assert kafka_key is not None

        if kafka_key is not None and kafka_key in self.Keys:
            return event

        else:
            return None



import logging

from .fileabcsource import FileABCSource

#

L = logging.getLogger(__file__)

#


class FileLineSource(FileABCSource):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        **Parameters**

        app: Application
                Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_

        pipeline : Pipeline
                Name of the Pipeline

        id : ID, default = None

        config : JSON, default = None
                Configuration file with additional information

        """
        super().__init__(app, pipeline, id=id, config=config)

    async def read(self, filename, f):
        """
        Description:

        **Parameters**

        filename :

        f :

        """

        for line in f:
            await self.process(line, {"filename": filename})

            await self.simulate_event()


#


class FileMultiLineSource(FileABCSource):
    """
    Description: Read file line by line but try to join multi-line events by separator.
    Separator is a (fixed) pattern that should present at the begin of the line, if it is a new event.

    Example:
    <133>1 2018-03-24T02:37:01+00:00 machine program 22068 - Start of the multiline event
            2nd line of the event
            3rd line of the event
    <133>1 2018-03-24T02:37:01+00:00 machine program 22068 - New event

    The separatpr is '<' string in this case

    |

    """

    def __init__(self, app, pipeline, separator, id=None, config=None):
        """
        Description:

        **Parameters**

        app: Application
                Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_

        pipeline : Pipeline
                Name of the Pipeline

        separator :

        id : ID, default = None

        config : JSON, default = None
                Configuration file with additional information

        """
        super().__init__(app, pipeline, id=id, config=config)

        if isinstance(separator, str):
            separator = separator.encode("utf-8")

        self._separator = separator
        # TODO: self._max_latch_size = 10000

    async def read(self, filename, f):
        """
        Description:

        **Parameters**

        filename :

        f :

        """
        latch = None

        for line in f:
            if line.startswith(self._separator) and latch is not None:
                await self.process(latch)
                latch = line

            else:
                if latch is None:
                    latch = line
                else:
                    latch = latch + b"\n" + line

        await self.simulate_event()

        if latch is not None:
            await self.process(latch, {"filename": filename})



def vscode_secrets_fixme():
    from IPython import get_ipython

    ipy = get_ipython()
    if not ipy:
        return
    import re
    import os

    if "IPKernelApp" in ipy.config or "VSCODE_PID" in os.environ:
        uri = ipy.get_parent()["metadata"]["cellId"]
        match_ = re.match(r"vscode-notebook-cell://[^/]+(/.*?)/[^/]+\.ipynb", uri)
        if match_:
            cleaned_path = match_.group(1) + "/"
            os.chdir(cleaned_path)


try:
    vscode_secrets_fixme()
except Exception:
    pass

from .jupyter import (
    deploy,
    init_bitswan_jupyter,
    sample_events,
    register_app_post_init,
    register_connection,
    retrieve_sample_events,
    register_lookup,
    new_pipeline,
    end_pipeline,
    register_source,
    register_processor,
    register_generator,
    register_sink,
    step,
    async_step,
    App,
    bitswan_auto_pipeline,
    auto_pipeline,
    bitswan_test_mode,
    add_test_probe,
    bitswan_test_probes,
    bitswan_tested_pipelines,
)

__all__ = [
    "deploy",
    "init_bitswan_jupyter",
    "sample_events",
    "register_app_post_init",
    "register_connection",
    "retrieve_sample_events",
    "register_lookup",
    "new_pipeline",
    "end_pipeline",
    "register_source",
    "register_processor",
    "register_generator",
    "register_sink",
    "step",
    "async_step",
    "App",
    "auto_pipeline",
    "bitswan_auto_pipeline",
    "bitswan_test_mode",
    "add_test_probe",
    "bitswan_test_probes",
    "bitswan_tested_pipelines",
]



import logging
import asyncio

import bspump.asab as asab


L = logging.getLogger(__name__)


class Source(asab.Configurable):
    """
    Description:

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Set the initial ID, :meth:`Pipeline <bspump.Pipeline()>` and Task.

        **Parameters**

        app : Application
                        Name of an `Application` <https://asab.readthedocs.io/en/latest/asab/application.html#>`_ .

        pipeline : address of a pipeline
                        Name of a :meth:`Pipeline <bspump.Pipeline()>`.

        id : str, default None
                        Name of a the :meth:`Pipeline <bspump.Pipeline()>`.

        config : compatible config type , default None
                        Option for adding a configuration file.

        """
        super().__init__(
            "pipeline:{}:{}".format(
                pipeline.Id, id if id is not None else self.__class__.__name__
            ),
            config=config,
        )

        self.Id = id if id is not None else self.__class__.__name__
        self.Pipeline = pipeline

        self.Task = None  # Contains a main coroutine `main()` if Pipeline is started

        self.MQTTService = app.get_service("bspump.MQTTService")

        self.EventCount = 0
        self.EventsToPublish = 0

    async def process(self, event, context=None):
        """
        This method is used to emit event into a :meth:`Pipeline <bspump.Pipeline()>`.

        **Parameters**

        event: Data with time stamp stored in any data type, usually JSON.
                        Message or information that is passed to the method and emitted into a :meth:`Pipeline <bspump.Pipeline()>`.

        context : default None
                        Additional information.

        If there is an error in the processing of the event, the :meth:`Pipeline <bspump.Pipeline()>` is throttled by setting the error and the exception raised.

        :hint The source should catch this exception and fail gracefully.

        """
        # TODO: Remove this method completely, each source should call pipeline.process() method directly

        self.EventCount += 1
        if self.MQTTService and self.EventsToPublish > 0:
            self.MQTTService.publish_event(
                self.Pipeline.Id, self, event, self.EventsToPublish
            )
            self.EventsToPublish -= 1

        await self.Pipeline.process(event, context=context)

    def start(self, loop):
        """
        Starts the :meth:`Pipeline <bspump.Pipeline()>` through the _main method, but if main method is implemented
        it starts the coroutine using main method instead.

        **Parameters**

        loop : ?
                        Contains the coroutines.


        """
        if self.Task is not None:
            return

        async def _main():
            """
            Description:

            :return:
            """
            # This is to properly handle a lifecycle of the main method
            try:
                await self.main()
            except asyncio.CancelledError:
                pass
            except Exception as e:
                self.Pipeline.set_error(None, None, e)
                L.exception("Exception in the source '{}'".format(self.Id))

        self.Task = loop.create_task(_main())

    async def stop(self):
        """
        Stops the Source using self.Task. If the processes are not done it cancels them or raises an error.

        :return:
        """

        if self.Task is None:
            return  # Source is not started

        if not self.Task.done():
            self.Task.cancel()

        try:
            await self.Task
        except asyncio.exceptions.CancelledError:
            L.warning("Task cancelled: {}".format(self))

        if not self.Task.done():
            L.error("Source '{}' refused to stop: {}".format(self.Id, self.Task))

        self.Task = None

    def restart(self, loop):
        """
        Restarts the loop of coroutines and returns result() method.

        **Parameters**

        loop : ??
                        Contains the coroutines.

        """
        if self.Task is not None:
            if self.Task.done():
                self.Task.result()
                self.Task = None
        self.start(loop)

    async def main(self):
        """
        Can be implemented for additional features, else will raise NotImplementedError and _main is called instead.

        """
        raise NotImplementedError()

    async def stopped(self):
        """
        Waits for all asynchronous tasks to be completed. It is helper that simplifies the implementation of sources.

        Example:

        ..code:: python

                        async def main(self):

                                        #... initialize resources here

                                        await self.stopped()

                                        #... finalize resources here

        """

        try:
            while True:
                await asyncio.sleep(60)

        except asyncio.CancelledError:
            pass

    def locate_address(self):
        """
        Locates address of a :meth:`Pipeline <bspump.Pipeline()>`.

        :return: ID and ID of a :meth:`Pipeline <bspump.Pipeline()>` as a string.

        """
        return "{}.*{}".format(self.Pipeline.Id, self.Id)

    def rest_get(self):
        """
        :return: ID and class ID

        """
        return {"Id": self.Id, "Class": self.__class__.__name__}

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.locate_address())

    @classmethod
    def construct(cls, app, pipeline, definition: dict):
        """
        Can create a source based on a specific definition. For example, a JSON file.

        **Parameters**

        app : Application
                        Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html#>`_.

        pipeline : :meth:`Pipeline <bspump.Pipeline()>`
                        Specification of a :meth:`Pipeline <bspump.Pipeline()>`.

        definition : dict
                        Definition that is used to create a source.


        :return: cls(app, newid, config)

        """
        newid = definition.get("id")
        config = definition.get("config")
        args = definition.get("args")
        if args is not None:
            return cls(app, pipeline, id=newid, config=config, **args)
        else:
            return cls(app, pipeline, id=newid, config=config)


class TriggerSource(Source):
    """
    A specialized source that responds to trigger events.

    TriggerSource is designed to work with BitSwan's trigger system. It waits
    for trigger events and executes a cycle() method when triggered. This allows
    for event-driven data processing in pipelines.

    Key features:
    - Waits for trigger events using asyncio.Event
    - Executes cycle() method when triggered
    - Manages multiple triggers simultaneously
    - Handles trigger lifecycle events

    Usage:
        class MyTriggerSource(TriggerSource):
            async def cycle(self, *args, **kwargs):
                # Your trigger logic here
                await self.Pipeline.process({"data": "triggered"})

        # Attach to any trigger type
        source = MyTriggerSource(app, pipeline)
        source.on(some_trigger)
    """

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.App = app
        self.Loop = app.Loop

        self.TriggerEvent = asyncio.Event()
        self.TriggerEvent.clear()
        self.Triggers = set()

    def time(self):
        """
        Method used for measuring an accurate time.

        :return: App.time()

        :hint: You can find more information about `UTC Time <https://asab.readthedocs.io/en/latest/asab/application.html#utc-time>`_ in the ASAB documentation

        """
        return self.App.time()

    def on(self, trigger):
        """
        Sets a Trigger which is a method that waits for a given condition.

        **Parameters**

        trigger : keyword of a trigger
                        Given condition that.

        :return: Trigger.add(trigger)

        """
        trigger.add(self)
        self.Triggers.add(trigger)
        return self

    async def main(self, *args, **kwags):
        """
        Waits for :meth:`Pipeline <bspump.Pipeline()>`, triggers, and calls exceptions when the source is initiated.

        **Parameters**

        *args : ?

        **kwags : ?

        |

        """
        while True:
            # Wait for pipeline is ready
            await self.Pipeline.ready()

            # Wait for a trigger
            try:
                await self.TriggerEvent.wait()
            except asyncio.CancelledError:
                break

            # Send begin on a cycle event
            self.Pipeline.PubSub.publish(
                "bspump.pipeline.cycle_begin!", pipeline=self.Pipeline
            )

            # Execute one cycle
            try:
                await self.cycle(*args, **kwags)

            except asyncio.CancelledError:
                # This happens when Ctrl-C is pressed
                L.warning(
                    "Pipeline '{}' processing was cancelled".format(self.Pipeline.Id)
                )

                # Send end of a cycle event
                self.Pipeline.PubSub.publish(
                    "bspump.pipeline.cycle_canceled!", pipeline=self.Pipeline
                )

                break

            except BaseException as e:
                self.Pipeline.set_error(None, None, e)

            # Send end of a cycle event
            self.Pipeline.PubSub.publish(
                "bspump.pipeline.cycle_end!", pipeline=self.Pipeline
            )

            self.TriggerEvent.clear()
            for trigger in self.Triggers:
                trigger.done(self)

    async def cycle(self, *args, **kwags):
        """
        Not implemented.

        **Parameters**

        *args : ?

        **kwags : ?


        """
        raise NotImplementedError()

    def rest_get(self):
        """
        Description:

        :return:
        """
        result = super().rest_get()
        result.update({"triggered": self.TriggerEvent.is_set()})
        return result



import logging
import asyncio
import json
import os
import pathlib

import jwt
from jwt.exceptions import ExpiredSignatureError, DecodeError
from typing import Callable

from .components import (
    BaseField,
    Field,
    FloatField,
    IntField,
    CheckboxField,
    TextField,
    ChoiceField,  # noqa: F401
    FieldSet,  # noqa: F401
    RawJSONField,  # noqa: F401
    FileField,  # noqa: F401
    Button,  # noqa: F401
    DateField,  # noqa: F401
    DateTimeField,  # noqa: F401
    TableComponent,  # noqa: F401
)


from .template_env import env

from ...abc.source import Source
from ...abc.sink import Sink
from ...abc.connection import Connection

import aiohttp.web
from aiohttp.web import Request
from importlib.resources import files

L = logging.getLogger(__name__)


def recursive_merge(dict1, dict2):
    for key, value in dict2.items():
        if key in dict1 and isinstance(dict1[key], dict) and isinstance(value, dict):
            dict1[key] = recursive_merge(dict1[key], value)
        else:
            dict1[key] = value
    return dict1


def expects_html(request):
    html = request.content_type == "text/html"
    accept = request.headers.get("Accept")
    if accept:
        if "text/html" in accept and request.content_type != "text/html":
            html = True
    return html


class WebServerConnection(Connection):
    """
    Source with events from a specific route.
    """

    ConfigDefaults = {
        "port": 8080,
        "max_body_size_bytes": 1024 * 1024 * 1000,
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        self.aiohttp_app = aiohttp.web.Application(
            client_max_size=int(self.Config["max_body_size_bytes"])
        )
        static_dir = str(files("bspump").joinpath("styles"))
        self.aiohttp_app.router.add_static("/styles/", static_dir, show_index=True)

        notebook_dir = os.path.dirname(app.Notebook)
        scripts_dir = str(
            pathlib.Path().resolve().joinpath(notebook_dir).joinpath("scripts")
        )
        if pathlib.Path(scripts_dir).exists():
            self.aiohttp_app.router.add_static(
                "/scripts/", scripts_dir, show_index=True
            )
        self.start_server()

    def start_server(self):
        print("Starting webserver")
        try:
            self.App.Loop.create_task(
                aiohttp.web._run_app(
                    self.aiohttp_app,
                    port=int(self.Config["port"]),
                )
            )
        except Exception as e:
            print("Exception: {}".format(e))
            import traceback

            traceback.print_exc()


class WebRouteSource(Source):
    """
    WebSource is a source that listens on a specified port and serves HTTP requests.
    """

    def __init__(
        self,
        app,
        pipeline,
        connection="DefaultWebServerConnection",
        method="GET",
        route="/",
        id=None,
        config=None,
    ):
        super().__init__(app, pipeline, id=id, config=config)
        pipeline.StopOnErrors = False

        try:
            self.Connection = pipeline.locate_connection(app, connection)
        except KeyError:
            if connection == "DefaultWebServerConnection":
                try:
                    default_port = int(
                        os.environ.get("DEFAULT_WEB_SERVER_CONNECTION_PORT") or "8080"
                    )
                except ValueError:
                    default_port = 8080
                    L.warning(
                        "DEFAULT_WEB_SERVER_CONNECTION_PORT is not a valid integer. Using default value {}.".format(
                            default_port
                        )
                    )

                self.Connection = WebServerConnection(
                    app,
                    "DefaultWebServerConnection",
                    {"port": default_port},
                )
                app.PumpService.add_connection(self.Connection)
        self.aiohttp_app = self.Connection.aiohttp_app
        self.aiohttp_app.router.add_route(method, route, self.handle_request)

    async def main(self):
        pass

    async def handle_request(self, request):
        try:
            response_future = asyncio.Future()
            await self.process(
                {
                    "request": request,
                    "response_future": response_future,
                    "status": 200,
                }
            )
            return await response_future
        except Exception:
            L.exception("Exception in WebSource")
            return aiohttp.web.Response(status=500)


async def gate_response(request, test_secret, response_fn):
    secret = request.query.get("secret")
    if secret is None:
        auth = request.headers.get("Authorization")
        if auth is not None:
            if auth.startswith("Bearer "):
                secret = auth[7:]
        if secret is None:
            return aiohttp.web.Response(
                text="Secret is missing. Pass via query parameter 'secret' or in the Authorization as 'Bearer <secret>'",
                status=401,
            )
    if not test_secret(secret):
        return aiohttp.web.Response(text="Invalid secret", status=403)
    return await response_fn()


class ProtectedWebRouteSource(WebRouteSource):
    """
    Web route source that requires a secret in a qparam or in the BearerToken.
    """

    async def handle_request(self, request: Request):
        try:

            async def response_fn():
                response_future = asyncio.Future()
                await self.process(
                    {
                        "request": request,
                        "response_future": response_future,
                        "status": 200,
                    }
                )
                return await response_future

            return await gate_response(
                request, lambda secret: self.test_secret(secret), response_fn
            )
        except Exception:
            L.exception("Exception in WebSource")
            return aiohttp.web.Response(status=500)

    def test_secret(self, secret):
        return secret == self.Config["secret"]


class WebFormSource(WebRouteSource):
    def __init__(
        self,
        app,
        pipeline,
        connection="DefaultWebServerConnection",
        route="/",
        fields: list[BaseField] | Callable[[Request], list[BaseField]] = lambda r: [],
        id=None,
        config=None,
        form_intro="",
    ):
        super().__init__(
            app,
            pipeline,
            connection=connection,
            route=route,
            method="GET",
            id=id,
            config=config,
        )
        self.fields = []
        self.generate_fields = None
        if isinstance(fields, list):
            self.fields = fields
        elif isinstance(fields, Callable):
            self.generate_fields = fields
        else:
            raise ValueError(
                f"incorrect type {type(fields)}. Expected list[Field] or Callable[[Request] -> list[Field]]"
            )

        self.form_intro = form_intro
        self.aiohttp_app.router.add_route("POST", route, self.handle_post)

    async def handle_request(self, request: Request):
        if self.generate_fields:
            self.fields = self.generate_fields(request)
        if request.content_type == "application/json":
            field_info = {f.name: f.get_params() for f in self.fields}
            return aiohttp.web.json_response(field_info)
        return aiohttp.web.Response(
            text=self.render_form(request),
            content_type="text/html",
        )

    def load_fields(self, request: Request):
        if self.generate_fields:
            self.fields = self.generate_fields(request)

    def validate_field_presence(self, data, fields=None):
        if not fields:
            fields = self.fields
        for field in fields:
            if field.required and field.name not in data:
                raise ValueError(f"Field {field.name} is required")

            if isinstance(field, FieldSet):
                if hasattr(field, "fields") and field.fields:
                    self.validate_field_presence(data[field.name], field.fields)
                elif field.required:
                    raise ValueError(f"Field {field.name} is required")

    async def extract_data(self, request: Request):
        if request.content_type == "application/json":
            data = await request.json()
        elif request.content_type == "application/x-www-form-urlencoded":
            data = await request.json()
        else:
            dfrom = dict(await request.post())
            data = {}
            for field in self.fields:
                field.restructure_data(dfrom, data)
        return data

    async def clean_and_process(self, request: Request, data):
        for field in self.fields:
            try:
                field.clean(data, request=request)
            except ValueError as e:
                if expects_html(request):
                    return aiohttp.web.Response(
                        text=self.render_form(request, errors={field.name: e}),
                        content_type="text/html",
                        status=400,
                    )
                else:
                    return aiohttp.web.json_response(
                        {"error": f"{field.name} is incorrectly formatted"}, status=400
                    )
        response_future = asyncio.Future()
        await self.process(
            {
                "request": request,
                "response_future": response_future,
                "status": 200,
                "form": data,
            }
        )
        return await response_future

    async def handle_post(self, request: Request):
        try:
            data = await self.extract_data(request)
        except json.decoder.JSONDecodeError:
            return aiohttp.web.Response(status=400, text="Invalid JSON")
        try:
            self.validate_field_presence(data)
        except ValueError as e:
            return aiohttp.web.Response(
                status=400,
                text=f"Schema validation error: {e}",
                content_type="text/plain",
            )
        return await self.clean_and_process(request, data)

    def extract_defaults(self, request: Request):
        defaults = {}
        for query_param, value in request.query.items():
            if "___" in query_param:
                parts = query_param.split("___")
                current_dict = defaults
                for fieldset in parts[:-1]:
                    if not current_dict.get(fieldset):
                        current_dict[fieldset] = {}
                    current_dict = current_dict[fieldset]
                current_dict[parts[-1]] = value
        return defaults

    def render_form(self, request: Request, errors={}):
        defaults = self.extract_defaults(request)

        for field in self.fields:
            if field.name in request.query:
                defaults[field.name] = request.query[field.name]
            elif field.name not in defaults:
                defaults[field.name] = field.default

        template = env.get_template("source-form.html")
        return template.render(
            form_intro=self.form_intro,
            fields=self.fields,
            defaults=defaults,
            errors=errors,
        )


class ProtectedWebFormSource(WebFormSource):
    async def handle_request(self, request: Request):
        su = super()

        async def response_fn():
            return await su.handle_request(request)

        return await gate_response(
            request, lambda secret: self.test_secret(secret), response_fn
        )

    def test_secret(self, secret):
        return secret == self.Config["secret"]

    async def handle_post(self, request: Request):
        su = super()

        async def response_fn():
            return await su.handle_post(request)

        return await gate_response(
            request, lambda secret: self.test_secret(secret), response_fn
        )


class JWTWebFormSource(ProtectedWebFormSource):
    """
    ProtectedWebFormSource that is gated by a JWT token issued at some time and
    due to expire. All hidden fields are encoded in the JWT token and values
    for any other fields encoded in the JWT token take precedence over user
    input
    """

    def __init__(
        self,
        app,
        pipeline,
        connection="DefaultWebServerConnection",
        route="/",
        fields: Field = None,
        id=None,
        config=None,
        form_intro="",
    ):
        self.fields = fields + [IntField("exp", hidden=True, display="Expires at: ")]
        super().__init__(
            app,
            pipeline,
            connection,
            route,
            self.fields,
            id=id,
            config=config,
            form_intro="",
        )

    def test_secret(self, secret):
        print(self.Config["jwt-secret"])
        try:
            jwt.decode(secret, self.Config["jwt-secret"], algorithms=["HS256"])
        except DecodeError as e:
            print(e)
            return False
        except ExpiredSignatureError as e:
            print(e)
            return False
        return True

    def extract_defaults(self, request: Request):
        su = super()
        defaults = su.extract_defaults(request)
        defaults = recursive_merge(
            defaults,
            jwt.decode(
                request.query["secret"], self.Config["jwt-secret"], algorithms=["HS256"]
            ),
        )
        return defaults

    async def handle_post(self, request: Request):
        try:
            data = await self.extract_data(request)
        except json.decoder.JSONDecodeError:
            return aiohttp.web.Response(status=400, text="Invalid JSON")

        try:
            data = recursive_merge(
                data,
                jwt.decode(
                    request.query["secret"],
                    self.Config["jwt-secret"],
                    algorithms=["HS256"],
                ),
            )
        except DecodeError as e:
            return aiohttp.web.Response(text=f"Invalid secret: {e}", status=400)
        except ExpiredSignatureError as e:
            return aiohttp.web.Response(text=f"JWT Token expired: {e}", status=403)

        try:
            self.validate_field_presence(data)
        except ValueError as e:
            return aiohttp.web.Response(
                status=400,
                text=f"Schema validation error: {e}",
                content_type="text/plain",
            )

        return await self.clean_and_process(request, data)


class WebSink(Sink):
    """
    WebSink is a sink that sends HTTP requests.
    """

    def process(self, context, event):
        content_type = event.get("content_type", "text/html")
        # if bytes we use a binary response otherwise we use text
        if isinstance(event["response"], bytes):
            event["response_future"].set_result(
                aiohttp.web.Response(
                    status=event["status"],
                    body=event["response"],
                    content_type=content_type,
                )
            )
        else:
            event["response_future"].set_result(
                aiohttp.web.Response(
                    status=event["status"],
                    text=event["response"],
                    content_type=content_type,
                )
            )


class RichJSONOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.buttons = []

    def add_button(self, name, on_click, button_id):
        self.buttons.append(Button(name=name, on_click=on_click, button_id=button_id))


class JSONWebSink(Sink):
    """
    JSONWebSink is a sink that sends HTTP requests with JSON content.
    """

    def send_response(self, event):
        if expects_html(event["request"]):
            html_content = self.render_html_output(event["response"])
            event["response_future"].set_result(
                aiohttp.web.Response(
                    text=html_content,
                    content_type="text/html",
                    status=event["status"],
                )
            )
        else:
            event["response_future"].set_result(
                aiohttp.web.json_response(event["response"], status=event["status"])
            )

    def process(self, context, event):
        """
        Process the incoming event and respond with either JSON or HTML.
        """
        self.send_response(event)

    def handle_error(self, context, event, exception, timestamp):
        """
        If an exception is raised in the pipeline step, return 500 and
        a page for the user
        """
        event["status"] = 500
        event["response"] = {"error": "Internal Server Error"}
        self.send_response(event)

    def render_html_output(self, json_data):
        template = env.get_template("output-form.html")
        fields_html = self.format_json_to_html(json_data)
        return template.render(fields=fields_html)

    def format_json_to_html(self, json_data):
        fields_html = []

        if hasattr(json_data, "buttons"):
            for button in json_data.buttons:
                fields_html.append(button.html())

        for key, value in json_data.items():
            if isinstance(value, dict):
                nested_html = self.format_json_to_html(value)
                template = env.get_template("nested-json.html")
                fields_html.append(template.render(key=key, content=nested_html))
            elif isinstance(value, list):
                nested_html = self.format_list(key, value)
                template = env.get_template("list.html")
                fields_html.append(template.render(key=key, content=nested_html))
            else:
                fields_html.append(self.format_key_value(key, value))

        return "".join(fields_html)

    def format_list(self, key, json_data_lst):
        list_items_html = []

        for item in json_data_lst:
            if isinstance(item, list):
                list_items_html.append(self.format_list(key, item))
            elif isinstance(item, dict):
                list_items_html.append(self.format_json_to_html(item))
            else:
                list_items_html.append(self.format_key_value(key, item))

        return "".join(list_items_html)

    def format_key_value(self, key, value):
        if isinstance(value, bool):
            fd = CheckboxField(key, readonly=True, default=value)
        elif isinstance(value, int):
            fd = IntField(key, readonly=True, default=value)
        elif isinstance(value, float):
            fd = FloatField(key, readonly=True, default=value)
        else:
            fd = TextField(key, readonly=True, default=value)
        return fd.html()



from ...abc import Expression


class UPPER(Expression):
    Attributes = {
        "Value": ["str"],
    }

    Category = "String"

    def __init__(self, app, *, arg_what):
        super().__init__(app)
        self.Value = arg_what

    def get_outlet_type(self):
        return str.__name__

    def consult_inlet_type(self, key, child):
        return str.__name__

    def __call__(self, context, event, *args, **kwargs):
        try:
            return self.Value(context, event, *args, **kwargs).upper()
        except AttributeError:
            return None



import abc
import asyncio
import collections
import concurrent
import datetime
import logging
import uuid
import traceback

import time

import bspump.asab
from bspump.asab.alert import Alert
from .abc.connection import Connection
from .abc.generator import Generator
from .abc.sink import Sink
from .abc.source import Source
from .analyzer import Analyzer
from .exception import ProcessingError

#

L = logging.getLogger(__name__)

#


class Pipeline(abc.ABC, bspump.asab.Configurable):
    """
    Description: Pipeline is ...

    An example of The :meth:`Pipeline <bspump.Pipeline()>` construction:

    .. code:: python

    class MyPipeline(bspump.Pipeline):

                    def __init__(self, app, pipeline_id):
                                    super().__init__(app, pipeline_id)
                                    self.build(
                                                    [
                                                    MySource(app, self),
                                                    MyProcessor(app, self),
                                                    MyProcessor2(app, self),
                                                    ]
                                            bspump.common.NullSink(app, self),
                                    )

    """

    ConfigDefaults = {
        "async_concurency_limit": 1000,  # TODO concurrency
        "reset_profiler": True,
        "stop_on_errors": True,
    }

    def __init__(self, app, id=None, config=None):
        """
        Initializes basic variables used in the other :meth:`Pipeline <bspump.Pipeline()>` methods. You can also add more information using parameters.

        **Parameters**

                        app : Application
                                        name of the ASAB `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_

                        id : str, default None
                                        You can enter ID of the class. Otherwise a name of the current class will used by calling __class__ descriptor object.

                        config : ,default None
                                        You can add a config file with additional settings and configurations, otherwise a default config is used.

        """
        _id = id if id is not None else self.__class__.__name__
        super().__init__("pipeline:{}".format(_id), config=config)

        self.Id = _id
        self.App = app
        self.Loop = app.Loop

        self.AlertService = app.AlertService

        self.MQTTService = app.get_service("bspump.MQTTService")
        self.PublishingProcessors = {}
        self.StopOnErrors = self.Config["stop_on_errors"]

        # Ensuring the uniqueness of the alert for each pipeline
        self.Alert_id = self.Config.get("alert_id", str(uuid.uuid4()))

        self.AsyncFutures = []
        self.AsyncConcurencyLimit = int(self.Config["async_concurency_limit"])
        self.ResetProfiler = self.Config.getboolean("reset_profiler")
        assert self.AsyncConcurencyLimit > 1

        # This object serves to identify the throttler, because list cannot be used as a throttler
        self.AsyncFuturesThrottler = object()

        self.Sources = []
        self.Processors = [
            []
        ]  # List of lists of processors, the depth is increased by a Generator object
        self.Sinks = []

        # Publish-Subscribe for this pipeline
        self.PubSub = bspump.asab.PubSub(app)
        self.MetricsService = app.get_service("asab.MetricsService")
        self.MetricsCounter = self.MetricsService.create_counter(
            "bspump.pipeline",
            tags={"pipeline": self.Id},
            init_values={
                "event.in": 0,
                "event.out": 0,
                "event.drop": 0,
                "warning": 0,
                "error": 0,
            },
        )
        self.MetricsEPSCounter = self.create_eps_counter()

        self.MetricsGauge = self.MetricsService.create_gauge(
            "bspump.pipeline.gauge",
            tags={"pipeline": self.Id},
            init_values={
                "warning.ratio": 0.0,
                "error.ratio": 0.0,
            },
        )
        self.MetricsDutyCycle = self.MetricsService.create_duty_cycle(
            "bspump.pipeline.dutycycle",
            tags={"pipeline": self.Id},
            init_values={
                "ready": False,
            },
        )

        # Processor's metrics
        self.ProfilerCounter = {}
        self.ProcessorsEPSMetrics = {}
        self.ProcessorsCounter = {}

        app.PubSub.subscribe("Metrics.flush!", self._on_metrics_flush)

        # Pipeline logger
        self.L = PipelineLogger(
            "bspump.pipeline.{}".format(self.Id), self.MetricsCounter
        )

        self.LastReadyStateSwitch = self.Loop.time()

        self._error = None  # None if not in error state otherwise there is a tuple (context, event, exc, timestamp)

        self._throttles = set()
        self._ancestral_pipelines = set()

        self._ready = asyncio.Event()
        self._ready.clear()

        # Chillout is used to break a pipeline processing to smaller tasks that allows other event in event loop to be processed
        self._chillout_trigger = 10000
        self._chillout_counter = 0

        self._context = {}

    def time(self):
        """
        Returns correct time.

        :return: App.time()

        :hint: More information in the ASAB documentation in `UTC Time <https://asab.readthedocs.io/en/latest/asab/application.html#utc-time>`_.

        """
        return self.App.time()

    def get_throttles(self):
        """
        Returns components from :meth:`Pipeline <bspump.Pipeline()>` that are throttled.

        :return: self._throttles
                        Return list of throttles.

        Parameters: ---


        :return: xxxx
        """
        return self._throttles

    def _on_metrics_flush(self, event_type):
        """
        Description: Pipeline is ...


        Parameters: event_type


        :return: xxxx
        """
        for field in self.MetricsCounter.Storage["fieldset"]:
            values = field["values"]
            if values["event.in"] == 0:
                self.MetricsGauge.set("warning.ratio", 0.0)
                self.MetricsGauge.set("error.ratio", 0.0)
                continue
            self.MetricsGauge.set(
                "warning.ratio", values["warning"] / values["event.in"]
            )
            self.MetricsGauge.set("error.ratio", values["error"] / values["event.in"])

        for processor in self.ProcessorsCounter:
            for field in self.ProcessorsCounter[processor].Storage["fieldset"]:
                values = field["values"]
                self.ProcessorsEPSMetrics[processor].add(
                    "eps.in", round(values["event.in"] / 60, 3)
                )
                self.ProcessorsEPSMetrics[processor].add(
                    "eps.out", round(values["event.out"] / 60, 3)
                )

    def is_error(self):
        """
        Returns False when there is no error, otherwise it returns True.

        :return: self._error is not None.

        Parameters: ---


        :return: xxxx
        """
        return self._error is not None

    def set_error(self, context, event, exc):
        """
        When called with `exc is None`, it resets error (aka recovery).

        When called with exc, it sets exceptions for soft errors.

        **Parameters**

        context : type?
                        Context of an error.

        event : Data with time stamp stored in any data type usually is in JSON.
                        You can specify an event that is passed to the method.

        exc : Exception.
                        Python default exceptions.

        """
        if exc is None:
            # Reset branch
            if self._error is not None:
                self._error = None
                L.info("Error cleared at a pipeline '{}'".format(self.Id))

                for source in self.Sources:
                    source.restart(self.Loop)

                self.PubSub.publish("bspump.pipeline.clear_error!", pipeline=self)
                self._evaluate_ready()

        else:
            tb_text = "".join(traceback.format_exception(None, exc, exc.__traceback__))

            alert_data = {
                "exception": "{}: {}".format(exc.__class__.__name__, exc),
                "event": str(event),
                "traceback": tb_text,
            }
            print(
                f"""
            Event: {alert_data['event']}
            Traceback: {alert_data['traceback']}"""
            )
            # send alert
            self.App.AlertService.trigger(
                Alert(
                    source=self.App.__class__.__name__,
                    alert_cls=self.Id,
                    alert_id=self.Alert_id,
                    title="{}:{} ERROR".format(self.Id, self.Alert_id),
                    data=alert_data,
                    exception=exc,
                )
            )

            if self.handle_error(exc, context, event):
                self.MetricsEPSCounter.add("warning", 1)
                self.MetricsCounter.add("warning", 1)
                self.PubSub.publish("bspump.pipeline.warning!", pipeline=self)
                return
            else:
                self.MetricsEPSCounter.add("error", 1)
                self.MetricsCounter.add("error", 1)

            if self._error is not None:
                L.warning("Error on a pipeline is already set!")

            self._error = (context, event, exc, self.App.time())

            L.exception(
                "Pipeline '{}' stopped due to a processing error: {} ({})".format(
                    self.Id, exc, type(exc)
                )
            )

            self.PubSub.publish("bspump.pipeline.error!", pipeline=self)
            self._evaluate_ready()

    def handle_error(self, exception, context, event):
        """
        Used for setting up exceptions and conditions for errors. You can implement it to evaluate processing errors.

        **Parameters**

        exception : Exception
                        Used for setting up a custom Exception.

        context : information
                        Additional information can be passed.

        event : Data with time stamp stored in any data type, usually it is in JSON.
                        You can specify an event that is passed to the method.

        :return: False for hard errors (stop the :meth:`Pipeline <bspump.Pipeline()>` processing). True for soft errors that will be ignored.

        Example:

        .. code:: python

                        class SampleInternalPipeline(bspump.Pipeline):

                                        def __init__(self, app, pipeline_id):
                                                        super().__init__(app, pipeline_id)

                                                        self.build(
                                                                        bspump.common.InternalSource(app, self),
                                                                        bspump.common.JSONParserProcessor(app, self),
                                                                        bspump.common.PPrintSink(app, self)
                                                        )

                                        def handle_error(self, exception, context, event):
                                                        if isinstance(exception, json.decoder.JSONDecodeError):
                                                                        return True
                                                        return False

        |

        """

        return not self.StopOnErrors

    def link(self, ancestral_pipeline):
        """
        Links this :meth:`Pipeline <bspump.Pipeline()>` with an ancestral :meth:`Pipeline <bspump.Pipeline()>`.
        This is needed e. g. for a propagation of the throttling from child :meth:`Pipelines <bspump.Pipeline()>` back to their ancestors.
        If the child :meth:`Pipeline <bspump.Pipeline()>` uses InternalSource, it may become throttled because the internal queue is full. If so,
        the throttling is propagated to the ancestral :meth:`Pipeline <bspump.Pipeline()>`, so that its source may block incoming events until the
        internal queue is empty again.

        **Parameters**

        ancestral_pipeline : str
                        ID of a :meth:`Pipeline <bspump.Pipeline()>` that will be linked.

        """

        self._ancestral_pipelines.add(ancestral_pipeline)

    def unlink(self, ancestral_pipeline):
        """
        Unlinks an ancestral :meth:`Pipeline <bspump.Pipeline()>` from this :meth:`Pipeline <bspump.Pipeline()>`.

        **Parameters**

        ancestral_pipeline : str
                        ID of a ancestral :meth:`Pipeline <bspump.Pipeline()>` that will be unlinked.

        """

        self._ancestral_pipelines.remove(ancestral_pipeline)

    def throttle(self, who, enable=True):
        """
        Enables throttling method for a chosen :meth:`pipeline <bspump.Pipeline()>` and its ancestral :meth:`pipelines <bspump.Pipeline()>`,x if needed.


        **Parameters**

        who : ID of a :meth:`processor <bspump.Processor()>`.
                        Name of a :meth:`processor <bspump.Processor()>` that we want to throttle.

        enable : bool, defualt True
                        When True, content of argument 'who' is added to _throttles list.


        """
        # L.debug("Pipeline '{}' throttle {} by {}".format(self.Id, "enabled" if enable else "disabled", who))
        if enable:
            self._throttles.add(who)
        else:
            if who in self._throttles:
                try:
                    self._throttles.remove(who)
                except KeyError:
                    raise KeyError("'{}' not present among throttles".format(who))

        # Throttle primary pipelines, if there are any
        for ancestral_pipeline in self._ancestral_pipelines:
            ancestral_pipeline.throttle(who=who, enable=enable)

        self._evaluate_ready()

    def _evaluate_ready(self):
        """
        Description:

        :return:
        """
        orig_ready = self.is_ready()

        # Do we observed an error?
        new_ready = self._error is None

        # Are we throttled?
        if new_ready:
            new_ready = len(self._throttles) == 0

        if orig_ready != new_ready:
            if new_ready:
                self._ready.set()
                self.PubSub.publish("bspump.pipeline.ready!", pipeline=self)
                self.MetricsDutyCycle.set("ready", True)
            else:
                self._ready.clear()
                self.PubSub.publish("bspump.pipeline.not_ready!", pipeline=self)
                self.MetricsDutyCycle.set("ready", False)

    async def ready(self):
        """
        Checks if the :meth:`Pipeline <bspump.Pipeline()>` is ready. The method can be used in source: `await self.Pipeline.ready()`.

        """

        self._chillout_counter += 1
        if self._chillout_counter >= self._chillout_trigger:
            self._chillout_counter = 0
            await asyncio.sleep(0)

        await self._ready.wait()
        return True

    def is_ready(self):
        """
        This method is a check up of the event in the Event class.

        :return: _ready.is_set().

        """
        return self._ready.is_set()

    def _do_process(self, event, depth, context):
        """
        Description:

        :return:
        """
        for processor in self.Processors[depth]:
            t0 = time.perf_counter()
            try:
                self.ProcessorsCounter[processor.Id].add("event.in", 1)
                event = processor.process(context, event)
                processor.EventCount += 1
                if (
                    self.MQTTService
                    and processor.Id in self.PublishingProcessors.keys()
                    and self.PublishingProcessors[processor.Id] > 0
                ):
                    self.MQTTService.publish_event(
                        self.Id,
                        processor,
                        event,
                        self.PublishingProcessors[processor.Id],
                    )
                    self.PublishingProcessors[processor.Id] -= 1
            except SystemExit as e:
                raise e
            except BaseException as e:
                self.ProcessorsCounter[processor.Id].add("event.drop", 1)
                if depth > 0:
                    raise e  # Handle error on the top depth
                self.set_error(context, event, e)
                event = None  # Event is discarted
            finally:
                self.ProcessorsCounter[processor.Id].add("event.out", 1)
                self.ProfilerCounter[processor.Id].add(
                    "duration", time.perf_counter() - t0
                )
                self.ProfilerCounter[processor.Id].add("run", 1)

            if event is None:  # Event has been consumed on the way
                if len(self.Processors) == (depth + 1):
                    if isinstance(processor, Sink):
                        # self.ProcessorsCounter[processor.Id].add('event.out', 1)
                        self.MetricsEPSCounter.add("eps.out", 1)
                        self.MetricsCounter.add("event.out", 1)
                    else:
                        self.ProcessorsCounter[processor.Id].add("event.drop", 1)
                        self.MetricsEPSCounter.add("eps.drop", 1)
                        self.MetricsCounter.add("event.drop", 1)
                return

        # NOTE The sink does not come up in self.Sinks. What was this supposed to do?
        if self.Sinks:
            for c, s in self.Sinks:
                if c(event):
                    e = s.process(context, event)
                    if e is not None:
                        event = e
                        break
            else:
                event = None
                return

        assert event is not None

        self.set_error(
            context,
            event,
            ProcessingError(
                "Incomplete pipeline, event '{}' is not consumed by a Sink".format(
                    event
                )
            ),
        )

    def inject(self, context, event, depth):
        """
        Injects method serves to inject events into the :meth:`Pipeline <bspump.Pipeline()>`'s depth defined by the depth attribute.
        Every depth is interconnected with a generator object.

        **Parameters**

        context : string
                        Information propagated through the :meth:`Pipeline <bspump.Pipeline()>`.

        event : Data with time stamp stored in any data type, usually it is in JSON.
                        You can specify an event that is passed to the method.

        depth : int
                        Level of depth.

        :note: For normal operations, it is highly recommended to use process method instead.

        """
        if context is None:
            context = self._context.copy()
        else:
            context = context.copy()
            context.update(self._context)

        self._error = (context, event, None, self.App.time())
        self._do_process(event, depth, context)

    async def process(self, event, context=None):
        """
        Process method serves to inject events into the :meth:`Pipeline <bspump.Pipeline()>`'s depth 0,
        while incrementing the event in metric.

        **Parameters**

        event : Data with time stamp stored in any data type, usually it is in JSON.
                        You can specify an event that is passed to the method.

        context : str, default None
                        You can add additional information needed for work with event streaming.

        :hint: This is recommended way of inserting events into a :meth:`Pipeline <bspump.Pipeline()>`.

        """

        while not self.is_ready():
            await self.ready()

        self.MetricsEPSCounter.add("eps.in", 1)
        self.MetricsCounter.add("event.in", 1)

        self.inject(context, event, depth=0)

    def create_eps_counter(self):
        """
        Creates a dictionary with information about the :meth:`Pipeline <bspump.Pipeline()>`. It contains eps (events per second), warnings and errors.

        :return: self.MetricsService
                        Creates eps counter using MetricsService.

        :note: EPS counter can be created using this method or dicertly by using MatricsService method.

        """
        return self.MetricsService.create_eps_counter(
            "bspump.pipeline.eps",
            tags={"pipeline": self.Id},
            init_values={
                "eps.in": 0,
                "eps.out": 0,
                "eps.drop": 0,
                "warning": 0,
                "error": 0,
            },
        )

    # Future methods

    def ensure_future(self, coro):
        """
        You can use this method to schedule a future task that will be executed in a context of the :meth:`Pipeline <bspump.Pipeline()>`.
        The :meth:`Pipeline <bspump.Pipeline()>` also manages a whole lifecycle of the future/task, which means,
        it will collect the future result, trash it, and mainly it will capture any possible exception,
        which will then block the :meth:`Pipeline <bspump.Pipeline()>` via set_error().

        **Parameters**

        coro : ??
                        ??

        :hint: If the number of futures exceeds the configured limit, the :meth:`Pipeline <bspump.Pipeline()>` is throttled.

        |

        """

        future = asyncio.ensure_future(coro)
        future.add_done_callback(self._future_done)
        self.AsyncFutures.append(future)

        # Throttle when the number of futures exceeds the max count
        if len(self.AsyncFutures) == self.AsyncConcurencyLimit:
            self.throttle(self.AsyncFuturesThrottler, True)

    def _future_done(self, future):
        # Remove the throttle
        if len(self.AsyncFutures) == self.AsyncConcurencyLimit:
            self.throttle(self.AsyncFuturesThrottler, False)

        self.AsyncFutures.remove(future)

        exception = future.exception()
        if exception is not None:
            try:
                *_, sink = self.iter_processors()
                context, event, _, timestamp = self._error
                event = sink.handle_error(context, event, exception, timestamp)
                self.set_error(context, event, exception)
                self._error = None
            except Exception as e:
                self.set_error(None, None, e)

    # Construction

    def set_source(self, source):
        """
        Sets a specific source or list of sources to the :meth:`Pipeline <bspump.Pipeline()>`.

        **Parameters**

        source : str, list optional
                        ID of a source.

        If a list of sources is passed to the method, it adds the entire list of sources to the :meth:`Pipeline <bspump.Pipeline()>`.
        """
        if isinstance(source, Source):
            self.Sources.append(source)
        else:
            self.Sources.extend(source)

    def append_processor(self, processor):
        """
        Adds a :meth:`Processors <bspump.Processor()>` to the current :meth:`Pipeline <bspump.Pipeline()>`.

        **Parameters**

        processor : str
                        ID of a :meth:`processor <bspump.Processor()>`.

        :hint: The Generator can be added by using this method. It requires a depth parameter.

        """
        for depth in self.Processors:
            if len(depth) != 0 and isinstance(depth[-1], Sink):
                L.exception("Cannot add {} after {}.".format(processor, depth[-1]))
        # TODO: Check if fitting
        self.Processors[-1].append(processor)

        if isinstance(processor, Generator):
            processor.set_depth(len(self.Processors) - 1)
            self.Processors.append([])

        self._post_add_processor(processor)

    def remove_processor(self, processor_id):
        """
        Removes a specific :meth:`processor <bspump.Processor()>` from the :meth:`Pipeline <bspump.Pipeline()>`.

        **Parameters**

        processor_id : str
                        ID of a :meth:`processor <bspump.Processor()>`.

        :raises: Error when :meth:`processor <bspump.Processor()>` is not found.

        """
        for depth in self.Processors:
            for idx, processor in enumerate(depth):
                if processor.Id != processor_id:
                    continue
                del depth[idx]
                del self.ProfilerCounter[processor.Id]
                del self.ProcessorsEPSMetrics[processor.Id]
                if isinstance(processor, Analyzer):
                    del self.ProfilerCounter["analyzer_" + processor.Id]
                return
        raise KeyError("Cannot find processor '{}'".format(processor_id))

    def insert_before(self, id, processor):
        """
        Inserts the :meth:`Processor <bspump.Processor()>` into the :meth:`Pipeline <bspump.Pipeline()>` in front of another :meth:`processor <bspump.Processor()>` specified by ID.

        **Parameters**

        id : str
                        ID of a :meth:`processor <bspump.Processor()>` that we want to insert.

        processor : str
                        Name of the :meth:`processor <bspump.Processor()>` in front of which will be inserted the new :meth:`processor <bspump.Processor()>`.

        :return: True on success. False if ID was not found.

        """
        for processors in self.Processors:
            for idx, _processor in enumerate(processors):
                if _processor.Id == id:
                    processors.insert(idx, processor)
                    self._post_add_processor(processor)
                    return True
        return False

    def insert_after(self, id, processor):
        """
        Inserts the :meth:`Processor <bspump.Processor()>` into the :meth:`Pipeline <bspump.Pipeline()>` behind another :meth:`Processors <bspump.Processor()>` specified by ID.

        **Parameters**

        id : str
                        ID of a processor that we want to insert.

        processor : str
                        Name of a :meth:`processor <bspump.Processor()>` after which we insert our :meth:`processor <bspump.Processor()>`.

        :return: True if successful. False if ID was not found.

        """
        for processors in self.Processors:
            for idx, _processor in enumerate(processors):
                if _processor.Id == id:
                    processors.insert(idx + 1, processor)
                    self._post_add_processor(processor)
                    return True
        return False

    def _post_add_processor(self, processor):
        """
        Description:

        :return:
        """
        self.ProfilerCounter[processor.Id] = self.MetricsService.create_counter(
            "bspump.pipeline.profiler",
            tags={
                "processor": processor.Id,
                "pipeline": self.Id,
            },
            init_values={"duration": 0.0, "run": 0},
            reset=self.ResetProfiler,
        )
        self.ProcessorsEPSMetrics[processor.Id] = self.MetricsService.create_counter(
            "bspump.pipeline.eps_processor",
            tags={
                "processor": processor.Id,
                "pipeline": self.Id,
            },
            init_values={
                "eps.in": 0,
                "eps.out": 0,
                "eps.drop": 0,
            },
        )

        self.ProcessorsCounter[processor.Id] = self.MetricsService.create_counter(
            "bspump.pipeline.processor",
            tags={
                "processor": processor.Id,
                "pipeline": self.Id,
            },
            init_values={
                "event.in": 0,
                "event.out": 0,
                "event.drop": 0,
            },
        )

        if isinstance(processor, Analyzer):
            self.ProfilerCounter[
                "analyzer_" + processor.Id
            ] = self.MetricsService.create_counter(
                "bspump.pipeline.profiler",
                tags={
                    "analyzer": processor.Id,
                    "pipeline": self.Id,
                },
                init_values={"duration": 0.0, "run": 0},
                reset=self.ResetProfiler,
            )

    def build(self, source, *processors):
        """
        This method enables to add sources, :meth:`Processors <bspump.Processor()>`, and sink to create the structure of the :meth:`Pipeline <bspump.Pipeline()>`.

        **Parameters**

        source : str
                        ID of a source.

        *processors : str, list optional
                        ID of :meth:`Processor <bspump.Processor()>` or list of IDs.

        """
        self.set_source(source)
        for processor in processors:
            if isinstance(processor, list):
                for p in processor:
                    self.Sinks.append((p[0], p[1]))
            else:
                self.append_processor(processor)

    def iter_processors(self):
        """
        Uses python generator routine that iterates through all :meth:`Processors <bspump.Processor()>` in the :meth:`Pipeline <bspump.Pipeline()>`.

        :yields: A Processor from a list in the :meth:`Pipeline <bspump.Pipeline()>`.

        """
        for processors in self.Processors:
            for processor in processors:
                yield processor

    # Locate  ...

    def locate_source(self, address):
        """
        Locates a sources based on its ID.

        **Parameters**

        address : str
                        ID of the source.

        """
        for source in self.Sources:
            if source.Id == address:
                return source
        return None

    def locate_connection(self, app, connection_id):
        """
        Finds a connection by ID.


        **Parameters**

        app : Application
                        Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_.

        connection_id : str
                        ID of connection we want to locate.

        :return: connection

        """
        if isinstance(connection_id, Connection):
            return connection_id
        svc = app.get_service("bspump.PumpService")
        connection = svc.locate_connection(connection_id)
        if connection is None:
            raise RuntimeError("Cannot locate connection '{}'".format(connection_id))
        return connection

    def locate_processor(self, processor_id):
        """
        Finds a :meth:`Processor <bspump.Processor()>` by ID.

        **Parameters**

        processor_id : str
                        ID of a :meth:`Processor <bspump.Processor()>`.

        :return: processor

        |

        """
        for processor in self.iter_processors():
            if processor.Id == processor_id:
                return processor

    # Lifecycle ...

    def start(self):
        """
        Starts the lifecycle of the :meth:`Pipeline <bspump.Pipeline()>`.

        """
        self.PubSub.publish("bspump.pipeline.start!", pipeline=self)

        # Start all non-started sources
        for source in self.Sources:
            source.start(self.Loop)

        self._evaluate_ready()

    async def stop(self):
        """
        Gracefully stops the lifecycle of the :meth:`Pipeline <bspump.Pipeline()>`.

        """
        self.PubSub.publish("bspump.pipeline.stop!", pipeline=self)

        # Stop all futures
        while len(self.AsyncFutures) > 0:
            # The futures are removed in _future_done
            await asyncio.wait(
                self.AsyncFutures, return_when=concurrent.futures.ALL_COMPLETED
            )

        # Stop all started sources
        for source in self.Sources:
            await source.stop()

    # Rest API

    def rest_get(self):
        """
        Returns information about the status of the :meth:`Pipeline <bspump.Pipeline()>`:

        :return:
        """
        rest = {
            "Id": self.Id,
            "Ready": self.is_ready(),
            "Throttles": list(self._throttles),
            "Sources": self.Sources,
            "Processors": [],
            "Metrics": self.MetricsService.Storage.Metrics,
            "Log": [record.__dict__ for record in self.L.Deque],
        }

        for processors in self.Processors:
            rest["Processors"].append(processors)

        if self._error:
            error_text = str(self._error[2])  # (context, event, exc, timestamp)[2]
            error_time = self._error[3]
            if len(error_text) == 0:
                error_text = str(type(self._error[2]))
            rest["Error"] = error_text
            rest["ErrorTimestamp"] = error_time

        return rest


###


class PipelineLogger(logging.Logger):
    """
    PipelineLogger is a feature of BSPump which enables direct monitoring of a specific :meth:`Pipeline <bspump.Pipeline()>`.
    It offers an overview of errors, error handling, data in a given time with its timestamp.

    """

    def __init__(self, name, metrics_counter, level=logging.NOTSET):
        """
        Itialize a metrics counter.


        """
        super().__init__(name, level=level)
        self.Deque = collections.deque([], 50)
        self._metrics_counter = metrics_counter

    # TODO: configurable maxlen that is now 50 ^^
    # TODO: configurable log level (per pipeline, from its config)

    def handle(self, record):
        """
        Counts and add errors to the error counter.

        **Parameters**

        record :
                        Record that is evaluated.


        """
        # Count errors and warnings
        if record.levelno == logging.WARNING:
            self._metrics_counter.add("warning", 1)
        elif record.levelno >= logging.ERROR:
            self._metrics_counter.add("error", 1)

        # Add formatted timestamp
        record.timestamp = self._format_time(record)

        # Add record
        self.Deque.append(record)

    def _format_time(self, record):
        """
        Description:

        :return:
        """
        ct = datetime.datetime.fromtimestamp(record.created)
        return ct.isoformat()



from .abc.service import Service


class StreamSocketServerService(Service):

    """
    Example of use:

    class ServiceMyProtocolServer(asab.StreamSocketServerService):

            async def initialize(self, app):
                    host = asab.Config.get('http', 'host')
                    port = asab.Config.getint('http', 'port')

                    L.debug("Starting server on {} {} ...".format(host, port))
                    await self.create_server(app, MyProtocol, [(host, port)])
    """

    def __init__(self, app, service_name):
        super().__init__(app, service_name)
        self._servers = []
        app.PubSub.subscribe("Application.exit!", self._on_exit)

    async def create_server(self, app, protocol, addrs):
        # TODO: Parallelize this ...
        for addr in addrs:
            host, port = addr
            server = await app.Loop.create_server(protocol, host, port)
            self._servers.append(server)

    def _on_exit(self, message_type):
        for server in self._servers:
            server.close()

    async def finalize(self, app):
        # TODO: Parallelize this ...
        for server in self._servers:
            await server.wait_closed()



import abc


class SessionStorage(abc.ABC):
    def __init__(self, app, max_age, session_class):
        self.App = app
        self.MaxAge = max_age
        self.SessionClass = session_class

    @abc.abstractmethod
    async def load(self, request):
        """
        Load a session object from a persistent store
        or create a new one using create() method, if needed
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def set(self, session, response):
        """
        Set session object reference to a response
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def delete(self, session):
        """
        Delete session object from a persistent store
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def store(self, session):
        """
        Store session object to persistent store (if needed)
        """
        raise NotImplementedError()

    async def create(self, request):
        return self.SessionClass(self, id=None, new=True, max_age=self.MaxAge)



from ..analyzer import TimeWindowAnalyzer
from .builder import ExpressionBuilder
from .abc import Expression, evaluate


class DeclarativeTimeWindowAnalyzer(TimeWindowAnalyzer):
    @classmethod
    def construct(cls, app, pipeline, definition: dict):
        _id = definition.get("id")
        config = definition.get("config")
        declaration = definition.get("declaration")
        return cls(app, pipeline, declaration=declaration, id=_id, config=config)

    def __init__(self, app, pipeline, declaration, id=None, config=None):
        builder = ExpressionBuilder(app)
        decl = builder.parse(declaration)
        define = decl.pop("define")
        self.Predicate = decl.pop("predicate")

        matrixes = define.pop("matrixes", {})
        dtype = matrixes.pop("primary", {}).get("cell", {})
        if isinstance(dtype, dict):
            dtype = {
                "names": list(dtype.keys()),
                "formats": list(dtype.values()),
            }

        columns = define.pop("columns", {})
        columns_count = columns.pop("count", 15)
        columns_resolution = columns.pop("resolution", 15)
        self.ColumnItem = columns.pop("item", "timestamp")

        self.Evaluates = decl.pop("evaluate")
        assert isinstance(self.Evaluates, dict)

        self.Dimension = define.pop("dimension")
        assert isinstance(self.Dimension, list) or isinstance(self.Dimension, str)

        self.Trigger = decl.pop("trigger", None)
        if self.Trigger is not None:
            assert isinstance(self.Trigger, list)

        # TODO: matrix_id
        # TODO: analyze_on_clock
        # TODO: start_time
        # TODO: clock_driven
        super().__init__(
            app,
            pipeline,
            id=id,
            config=config,
            dtype=dtype,
            columns=columns_count,
            resolution=columns_resolution,
        )

    def process(self, context, event):
        x = self.Predicate(context, event)
        if x:
            if x:
                x = "evaluate"
            if isinstance(x, list):
                for i in x:
                    evaluate(self.Evaluates[i], context, event)
            else:
                evaluate(self.Evaluates[x], context, event)

        return event

    def evaluate(self, evalobj, context, event):
        # Find the column in timewindow matrix to fit in
        column = event[self.ColumnItem]
        column = self.TimeWindow.get_column(column)
        if column is None:
            return

        # Find the row, existing or create a new one
        row_key = tuple(event[item] for item in self.Dimension)
        row = self.TimeWindow.get_row_index(row_key)
        if row is None:
            row = self.TimeWindow.add_row(row_key)

        cell = self.TimeWindow.Array[row, column]

        matrix = evalobj.get("primary")
        if isinstance(matrix, Expression):
            ret = evaluate(
                matrix,
                context,
                event,
                cell,
                row=row,
                column=column,
            )
        else:
            ret = matrix

        self.TimeWindow.Array[row, column] = ret
        self.trigger(context, event, "primary", ret, row, column)

    def trigger(self, context, event, matrix, cell, row, column):
        for t in self.Trigger:
            matrix = t[matrix]
            test = matrix["test"]
            result = evaluate(
                test,
                context,
                event,
                cell,
                row=row,
                column=column,
            )

            if result:
                print(">>> trigger", event, row, column)



from ...abc import Expression, evaluate


class FOR(Expression):
    Attributes = {
        "Each": ["*"],  # TODO: This ...
        "Do": ["*"],  # TODO: This ...
    }

    Category = "Statements"

    def __init__(self, app, *, arg_each, arg_do):
        super().__init__(app)
        self.Each = arg_each
        self.Do = arg_do

    def __call__(self, context, event, *args, **kwargs):
        return [
            evaluate(self.Do, context, event, item, *args, **kwargs)
            for item in evaluate(self.Each, context, event, *args, **kwargs)
        ]



import logging

from ..abc.source import TriggerSource

L = logging.getLogger(__name__)


class ElasticSearchSource(TriggerSource):
    """
    Description:

    """

    ConfigDefaults = {
        "index": "index-*",
        "scroll_timeout": "1m",
    }

    def __init__(
        self,
        app,
        pipeline,
        connection,
        request_body=None,
        paging=True,
        id=None,
        config=None,
    ):
        """

        **Parameters**

        app : Application
                        Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_.

        pipeline : Pipeline
                        Name of the Pipeline.

        connection : Connection
                        Information of the connection.

        request_body JSON, default = None
                        Request body needed for the request API call.

        paging : ?, default = True

        id : ID, default = None
                        ID

        config : JSON/dict, default = None
                        Configuration file with additional information.

        """
        super().__init__(app, pipeline, id=id, config=config)
        self.Connection = pipeline.locate_connection(app, connection)

        self.Index = self.Config["index"]
        self.ScrollTimeout = self.Config["scroll_timeout"]
        self.Paging = paging

        if request_body is not None:
            self.RequestBody = request_body
        else:
            self.RequestBody = {"query": {"bool": {"must": {"match_all": {}}}}}

    async def cycle(self):
        """
        Gets data from Elastic and injects them into the pipeline.

        """
        scroll_id = None

        while True:
            if scroll_id is None:
                path = "{}/_search?scroll={}".format(self.Index, self.ScrollTimeout)
                request_body = self.RequestBody
            else:
                path = "_search/scroll"
                request_body = {"scroll": self.ScrollTimeout, "scroll_id": scroll_id}

            url = self.Connection.get_url() + path
            async with self.Connection.get_session() as session:
                async with session.post(
                    url, json=request_body, headers={"Content-Type": "application/json"}
                ) as response:
                    if response.status != 200:
                        data = await response.text()
                        L.error(
                            "Failed to fetch data from ElasticSearch: {} from {}\n{}".format(
                                response.status, url, data
                            )
                        )
                        break

                    msg = await response.json()

            scroll_id = msg.get("_scroll_id")
            if scroll_id is None:
                break

            hits = msg["hits"]["hits"]
            if len(hits) == 0:
                break

            # Feed messages into a pipeline
            for hit in hits:
                await self.process(hit["_source"])

            if not self.Paging:
                break


class ElasticSearchAggsSource(TriggerSource):
    """
    Description:

    """

    ConfigDefaults = {
        "index": "index-*",
    }

    def __init__(
        self, app, pipeline, connection, request_body=None, id=None, config=None
    ):
        """
        Description:

        **Parameters**

        app : Application
                        Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_.

        pipeline : Pipeline
                        Name of the Pipeline.

        connection : Connection
                        Information of the connection.

        request_body JSON, default = None
                        Request body needed for the request API call.

        id : ID, default = None
                        ID info

        config : JSON/dict, default = None
                        configuration file with additional information.

        """
        super().__init__(app, pipeline, id=id, config=config)
        self.Connection = pipeline.locate_connection(app, connection)

        self.Index = self.Config["index"]

        if request_body is not None:
            self.RequestBody = request_body
        else:
            self.RequestBody = {"query": {"bool": {"must": {"match_all": {}}}}}

    async def cycle(self):
        """
        Sets request body and path to create query call.

        |

        """
        request_body = self.RequestBody
        path = "{}/_search?".format(self.Index)
        url = self.Connection.get_url() + path
        async with self.Connection.get_session() as session:
            async with session.post(
                url, json=request_body, headers={"Content-Type": "application/json"}
            ) as response:
                if response.status != 200:
                    data = await response.text()
                    L.error(
                        "Failed to fetch data from ElasticSearch: {} from {}\n{}".format(
                            response.status, url, data
                        )
                    )
                    return

                msg = await response.json()

        aggs = msg["aggregations"]

        if len(aggs) == 0:
            return

        start_name = list(aggs.keys())[0]
        start = aggs[start_name]

        path = {}
        await self.process_aggs(path, start_name, start)

    async def process_aggs(self, path, aggs_name, aggs):
        """
        Description:

        **Parameters**

        path :

        aggs_name :

        agss :

        """
        if "buckets" in aggs:
            await self.process_buckets(path, aggs_name, aggs["buckets"])

        if "value" in aggs:
            path[aggs_name] = aggs["value"]

            event = {}
            event.update(path)
            await self.process(event)
            path.pop(aggs_name)

    async def process_buckets(self, path, parent, buckets):
        """
        Recursive function for buckets processing.
        It iterates through keys of the dictionary, looking for 'buckets' or 'value'.
        If there are 'buckets', calls itself, if there is 'value', calls process_aggs
        and sends an event to process

        **Parameters**

        path :

        parent :

        buckets :

        """
        for bucket in buckets:
            for k in bucket.keys():
                if k == "key":
                    path[parent] = bucket[k]
                elif isinstance(bucket[k], dict):
                    await self.process_aggs(path, k, bucket[k])



from ..ipc import DatagramSource, StreamSource

# Backward compatibility
from ..ipc import DatagramSource as UDPSource
from ..ipc import StreamSource as TCPSource
from ..ipc import StreamSource as TCPStreamSource


__all__ = [
    "DatagramSource",
    "StreamSource",
    "UDPSource",
    "TCPSource",
    "TCPStreamSource",
]



import inspect

import bspump

from .builder import ExpressionBuilder


class DeclarativeGenerator(bspump.Generator):
    """
    Example of usage:

            ---
            pipeline_id: AnalyticsUserPipeline
            generator: |
                    --- !DICT
                    with: !EVENT
                    add:
                            user_password_info:
                                    !LOOKUP
                                    lookup: UserPasswordLookup (synchronous/asynchronous lookup)
                                    key:
                                            !ITEM
                                            with: !EVENT
                                            item: user
    """

    @classmethod
    def construct(cls, app, pipeline, definition: dict):
        _id = definition.get("id")
        config = definition.get("config")
        declaration = definition.get("declaration")
        return cls(app, pipeline, declaration=declaration, id=_id, config=config)

    @staticmethod
    async def _process_futures_in_dict(_dict):
        for field, value in _dict.items():
            if inspect.iscoroutine(value):
                _dict[field] = await value
        return _dict

    def __init__(self, app, pipeline, declaration, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        builder = ExpressionBuilder(app)
        self.Expression = builder.parse(declaration)

    async def generate(self, context, event, depth):
        self.Pipeline.inject(
            await self._process_futures_in_dict(context),
            await self._process_futures_in_dict(self.Expression(context, event)),
            depth,
        )



from ...abc import Expression
from ..value.valueexpr import VALUE


class STARTSWITH(Expression):
    Attributes = {
        "What": ["str"],
        "Prefix": ["str"],
    }

    Category = "String"

    def __init__(self, app, *, arg_what, arg_prefix):
        super().__init__(app)

        if isinstance(arg_what, Expression):
            self.What = arg_what
        else:
            self.What = VALUE(app, value=arg_what)

        if isinstance(arg_prefix, Expression):
            self.Prefix = arg_prefix
        else:
            self.Prefix = VALUE(app, value=arg_prefix)

    def __call__(self, context, event, *args, **kwargs):
        value = self.What(context, event, *args, **kwargs)
        if value is None:
            return False

        prefix = self.Prefix(context, event, *args, **kwargs)
        return value.startswith(prefix)

    def get_outlet_type(self):
        return bool.__name__

    def consult_inlet_type(self, key, child):
        return str.__name__



import logging

import numpy as np

from .analyzer import Analyzer

###

L = logging.getLogger(__name__)

###


class TimeDriftAnalyzer(Analyzer):
    """
    The analyzer, which shows how different is time of the stream from the current time.
    The output of the analyzis is a metric with average time, median time, minimum time,
    maximum time and a standart deviation.

    **Default Config**

    analyze_period : 5*60
            Once per 5 minutes.
    history_size : 100
            Keep maximum 100 array members.
    sparse_count : 1
            Process every single event.
    timestamp_attr : @timestamp
            Timestamp attribute present in the event to perform the drift analyzer on.
    """

    ConfigDefaults = {
        "analyze_period": 5 * 60,  # once per 5 minutes
        "history_size": 100,  # keep maximum 100 array members
        "sparse_count": 1,  # process every single event
        "timestamp_attr": "@timestamp",  # timestamp attribute present in the event to perform the drift analyzer on
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application.

        pipeline : Pipeline
                Name of the Pipeline.

        id : str, default = None
                ID

        config : JSON, default = None
                Configuration file with additional information.

        """
        # def __init__(self, app, pipeline, analyze_on_clock=False, analyze_period=None, id=None, config=None):
        super().__init__(app, pipeline, analyze_on_clock=True, id=id, config=config)

        self.History = []
        self.HistorySize = int(self.Config["history_size"])

        metrics_service = app.get_service("asab.MetricsService")

        self.DifferenceCounter = metrics_service.create_counter(
            "timedrift.difference", tags={}, init_values={"positive": 0, "negative": 0}
        )
        self.Gauge = metrics_service.create_gauge(
            "timedrift",
            tags={
                "pipeline": pipeline.Id,
            },
            init_values={
                "avg": 0.0,
                "median": 0.0,
                "stddev": 0.0,
                "min": 0.0,
                "max": 0.0,
            },
        )

        self.EventCount = 0
        self.SparseCount = int(self.Config["sparse_count"])

        self.TimestampAttr = self.Config["timestamp_attr"]

        self.App = app

    def predicate(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event : any data type
                information with timestamp

        :return: True

        """
        if self.TimestampAttr not in event:
            return False

        self.EventCount += 1
        if (self.EventCount % self.SparseCount) != 0:
            return False

        return True

    def get_diff(self, event_timestamp):
        """
        Returns the time difference of current event.

        **Parameters**

        event_timestamp : ?

        :return: diff
        """
        diff = self.App.time() - event_timestamp
        return diff

    def evaluate(self, context, event):
        """
        Description:

        """
        timestamp = event[self.TimestampAttr]
        diff = self.get_diff(timestamp)

        if diff < 0:
            self.DifferenceCounter.add("negative", 1)
            return
        self.DifferenceCounter.add("positive", 1)

        self.History.append(diff)

        while len(self.History) > self.HistorySize:
            self.History.pop(0)

    def analyze(self):
        """
        Description:

        """
        # in seconds
        if len(self.History) > 0:
            avg = np.mean(self.History)
            median = np.median(self.History)
            stddev = np.std(self.History)
            min_v = np.min(self.History)
            max_v = np.max(self.History)
            self.History = []
        else:
            avg = median = stddev = min_v = max_v = 0.0

        self.Gauge.set("avg", avg)
        self.Gauge.set("median", median)
        self.Gauge.set("stddev", stddev)
        self.Gauge.set("min", min_v)
        self.Gauge.set("max", max_v)



import os
import time

import bspump.declarative
import bspump.unittest


class TestDeclarativeTime(bspump.unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Builder = bspump.declarative.ExpressionBuilder(self.App)

    def load(self, decl_fname):
        basedir = os.path.dirname(__file__)
        with open(os.path.join(basedir, decl_fname), "r") as f:
            return self.Builder.parse(f.read())[0]

    def test_now_01(self):
        t1 = time.time()

        decl = self.load("./test_now.yaml")
        res = decl({}, {})

        self.assertTrue(res >= t1)
        self.assertTrue(res <= time.time())



import logging

from ...abc.source import Source

#

L = logging.getLogger(__name__)


def add_context_from_request(request, context):
    # Prepare context variables
    context["webservicesource.remote"] = request.remote
    context["webservicesource.path"] = request.path
    context["webservicesource.query"] = request.query
    context["webservicesource.headers"] = request.headers
    context["webservicesource.secure"] = request.secure
    context["webservicesource.method"] = request.method
    context["webservicesource.url"] = request.url
    context["webservicesource.scheme"] = request.scheme
    context["webservicesource.forwarded"] = request.forwarded
    context["webservicesource.host"] = request.host
    context["webservicesource.cookies"] = request.cookies
    context["webservicesource.content_type"] = request.content_type
    context["webservicesource.charset"] = request.charset
    return context


class WebServiceSource(Source):
    """
    This source is to be integrated into aiohttp.web as a 'View'.

    Example:
            async def view(self, request):
                    await self.WebServiceSource.put(None, data, request)
                    return aiohttp.web.Response(text='OK')

    """

    async def put(self, context, data, request):
        if context is None:
            context = {}

        add_context_from_request(request, context)
        await self.Pipeline.ready()
        await self.process(data, context)

    async def main(self):
        pass


class WebHookSource(Source):
    """
    Creates a webhook endpoint as configured in the configuration file.

    Configuration example:
    config = {
        "path": "/webhook",
        "port": 8080,
        "secret_qparam": "secret",
    }

    This will create a webhook endpoint at http://localhost:8080/webhook?secret=SECRET
    """

    ConfigDefaults = {
        "path": "webhook/",
        "port": "8080",
        "max_body_size_bytes": 1024 * 1000 * 1000,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Initializes parameters.

        **Parameters**

        app : Application
                Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html#>`_.

        pipeline : Pipeline
                Name of the Pipeline.

        id : , default = None

        config : , default = None

        """
        super().__init__(app, pipeline, id=id, config=config)

        self.App = app

    async def main(self):
        import aiohttp.web

        aiohttp_app = aiohttp.web.Application(
            client_max_size=self.Config["max_body_size_bytes"]
        )
        print("Adding routes to webserver")
        aiohttp_app.add_routes(
            [aiohttp.web.post(self.Config["path"], self.handle_post)]
        )
        print("Starting webserver")
        try:
            self.App.Loop.create_task(
                aiohttp.web._run_app(
                    aiohttp_app,
                    port=self.Config["port"],
                )
            )
        except Exception as e:
            print("Exception: {}".format(e))
            # print full traceback
            import traceback

            traceback.print_exc()

    async def handle_post(self, request):
        import aiohttp.web

        try:
            if self.Config.get("secret_qparam") is not None:
                secret = request.query.get("secret")
                if secret is None:
                    return aiohttp.web.Response(status=403, text="Missing secret")
                if secret != self.Config.get("secret_qparam"):
                    return aiohttp.web.Response(status=403, text="Invalid secret")
            body = await request.text()
            context = {}
            add_context_from_request(request, context)

            await self.process(body, context)
            return aiohttp.web.Response(text="Processed successfully")

        except Exception as e:
            L.exception("Error when processing POST request")
            self.Pipeline.set_error(None, None, e)
            return aiohttp.web.Response(text="Error processing request", status=500)



import bspump.crypto
import bspump.unittest


class CommonAESConfig(object):
    ConfigDefaults = {
        "key": "FD49F6AD21E24F36E51F7B49C37B1A97",
        "iv": "99C07A9AE2AA16486A21E504ECE4377D",
    }


class CustomEncryptAESProcessor(bspump.crypto.EncryptAESProcessor, CommonAESConfig):
    pass


class CustomDecryptAESProcessor(bspump.crypto.DecryptAESProcessor, CommonAESConfig):
    pass


class TestEncryptAESProcessor(bspump.unittest.ProcessorTestCase):
    def test_encrypt_aes_processor(self):
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(CustomEncryptAESProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][1]), 16)
        self.assertEqual(
            output[0][1], b"3\xeaj\xb6\xc4S\xc2\x8a\xef]\xf6\xfc\t\xc1\xd2\x02"
        )


class TestDecryptAESProcessor(bspump.unittest.ProcessorTestCase):
    def test_encrypt_aes_processor(self):
        events = [
            (None, b"3\xeaj\xb6\xc4S\xc2\x8a\xef]\xf6\xfc\t\xc1\xd2\x02"),
        ]

        self.set_up_processor(CustomDecryptAESProcessor)

        output = self.execute(events)

        self.assertEqual(output[0][1], b"deadc0de")



from abc import ABC, abstractmethod

from bspump import Generator


class AggregationStrategy(ABC):
    """
    Aggregation Strategy

    |

    """

    @abstractmethod
    def append(self, context, event):
        """
        Appends

        **Parameters**

        context :

        event :

        """
        raise NotImplementedError()

    @abstractmethod
    def flush(self):
        """
        Flushes

        |

        """
        raise NotImplementedError()

    @abstractmethod
    def is_empty(self) -> bool:
        """
        Description:

        |

        """
        raise NotImplementedError()


class ListAggregationStrategy(AggregationStrategy):
    """
    Description: ... test

    |

    """

    def __init__(self) -> None:
        """
        Description:

        |

        """
        super().__init__()
        self.AggregatedEvent = []

    def append(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event :

        """
        self.AggregatedEvent.append((context, event))

    def flush(self):
        """
        Description:

        :return: result

        |

        """
        result = self.AggregatedEvent
        self.AggregatedEvent = []
        return result

    def is_empty(self) -> bool:
        """
        Description:

        :return: Aggregated Event

        |

        """
        return len(self.AggregatedEvent) == 0


class ListEventAggregationStrategy(AggregationStrategy):
    """
    Description:

    |

    """

    def __init__(self) -> None:
        """
        Description:

        |

        """
        super().__init__()
        self.AggregatedEvent = []

    def append(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event :

        """
        self.AggregatedEvent.append(event)

    def flush(self):
        """
        Description:

        :return: result

        |

        """
        result = self.AggregatedEvent
        self.AggregatedEvent = []
        return result

    def is_empty(self) -> bool:
        """
        Description:

        :return: Aggregated event

        |

        """
        return len(self.AggregatedEvent) == 0


class StringAggregationStrategy(AggregationStrategy):
    """
    Description:

    |

    """

    def __init__(self, delimiter="\n") -> None:
        """
        Description:


        """
        super().__init__()
        self.Delimiter = delimiter
        self.AggregatedEvent = ""

    def append(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event : Data with time stamp stored in any data type usually is in JSON.
                You can specify an event that is passed to the method.

        """
        self.AggregatedEvent += str(event) + self.Delimiter

    def flush(self):
        """
        Description:

        :return: result

        |

        """
        result = self.AggregatedEvent[
            0 : -len(self.Delimiter)
        ]  # Remove trailing delimiter
        self.AggregatedEvent = ""
        return result

    def is_empty(self) -> bool:
        """
        Description:

        :return: Aggregated event

        |

        """
        return len(self.AggregatedEvent) == 0


class Aggregator(Generator):
    """
    Description:

    |

    """

    ConfigDefaults = {
        "completion_size": 10,
        "completion_timeout": 0,  # 0 means no timeout,
        "completion_interval": 0,  # 0 means no completion interval
    }

    def __init__(
        self,
        app,
        pipeline,
        aggregation_strategy: AggregationStrategy = ListAggregationStrategy(),
        id=None,
        config=None,
    ):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.CompletionSize = int(self.Config["completion_size"])
        self.CompletionTimeout = int(self.Config["completion_timeout"])
        self.CompletionInterval = int(self.Config["completion_interval"])

        if self.CompletionTimeout > 0 and self.CompletionInterval > 0:
            raise ValueError(
                "completion_timeout and completion_interval cannot be combined"
            )

        self.AggregationStrategy = aggregation_strategy

        self.CurrentSize = 0
        self.LastFlushTime = self.App.time()
        self.LastPeriodicFlushTime = self.App.time()

        app.PubSub.subscribe("Application.stop!", self._on_application_stop)

        if self.CompletionTimeout > 0:
            app.PubSub.subscribe("Application.tick!", self._check_timeout)

        if self.CompletionInterval > 0:
            app.PubSub.subscribe("Application.tick!", self._check_periodic_flush)

    def _check_timeout(self, _):
        if (
            self.CurrentSize > 0
            and self.App.time() - self.LastFlushTime > self.CompletionTimeout
        ):
            self.flush()

    def _check_periodic_flush(self, _):
        if (
            self.CurrentSize > 0
            and self.App.time() - self.LastPeriodicFlushTime > self.CompletionInterval
        ):
            self.LastPeriodicFlushTime = self.App.time()
            self.flush()

    def _on_application_stop(self, _, __):
        self.flush()

    def flush(self):
        """
        Description:

        :return: ??

        |

        """
        if self.AggregationStrategy.is_empty():
            return

        aggregated = self.AggregationStrategy.flush()
        self.Pipeline.ensure_future(
            self.generate({}, aggregated, self.PipelineDepth + 1)
        )

    def process(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event :

        """
        self.AggregationStrategy.append(context, event)
        self.CurrentSize += 1
        if self.CurrentSize >= self.CompletionSize:
            self.CurrentSize = 0
            self.flush()
        return None

    async def generate(self, context, aggregated_event, depth):
        """
        Description:

        **Parameters**

        context :

        aggregated_event :

        depth :

        """
        self.LastFlushTime = self.App.time()
        await self.Pipeline.inject(context, aggregated_event, depth)



from .matrixexportcsvgenerator import (
    SessionMatrixExportCSVGenerator,
    TimeWindowMatrixExportCSVGenerator,
)
from .source import MatrixSource
from .matrix import Matrix, PersistentMatrix
from .namedmatrix import NamedMatrix, PersistentNamedMatrix  # noqa: F401
from .timewindowmatrix import TimeWindowMatrix, PersistentTimeWindowMatrix  # noqa: F401
from .sessionmatrix import SessionMatrix, PersistentSessionMatrix
from .geomatrix import GeoMatrix, PersistentGeoMatrix


__all__ = [
    "SessionMatrixExportCSVGenerator",
    "TimeWindowMatrixExportCSVGenerator",
    "MatrixSource",
    "Matrix",
    "PersistentMatrix",
    "NamedMatrix",
    "PersistentNamedMatrix" "TimeWindowMatrix",
    "PersistentTimeWindowMatrix",
    "SessionMatrix",
    "PersistentSessionMatrix",
    "GeoMatrix",
    "PersistentGeoMatrix",
]



from .. import rest


class TenantWebHandler(object):
    def __init__(self, app, svc, web_container):
        self.TenantService = svc

        web_app = web_container.WebApp
        web_app.router.add_get("/tenant", self.get_tenants)

    async def get_tenants(self, request):
        tenants = self.TenantService.get_tenants()
        return rest.json_response(request, data=tenants)



from ..abc.sink import Sink


class NullSink(Sink):
    """
    Description:

    |

    """

    def process(self, context, event):
        """
        Description:

        |

        """
        pass



import logging
import asyncio
import aiohttp

from ..log import LOG_NOTICE

#

L = logging.getLogger(__name__)

#


class WebSocketFactory(object):

    """
    wsfactory = asab.web.WebSocketFactory(self)
    websvc.WebApp.router.add_get('/api/ws', wsfactory)
    wsfactory.on_message = ...


    Javascript example:

    <script type="text/javascript">
            var connection = new WebSocket(((window.location.protocol === "https:") ? "wss://" : "ws://") + window.location.host + "/api/ws");
            connection.onmessage = function (e) {
                    ...
            };
    </script>

    """

    def __init__(
        self, app, *, timeout=10.0, protocols=(), compress=True, max_msg_size=4194304
    ):
        self.App = app
        self.Counter = 0
        self.WebSockets = {}

        self.Timeout = timeout
        self.Protocols = protocols
        self.Compress = compress
        self.MaxMsgSize = max_msg_size
        self.__doc__ = "Provides WebSocket connection. Requires WebSocket upgrade."

        app.PubSub.subscribe("Application.stop!", self._on_app_stop)

    async def _on_app_stop(self, message_type, counter):
        # Clean up during application exit
        await self.close_all(
            code=aiohttp.WSCloseCode.GOING_AWAY, message="Server shutdown"
        )

    async def close_all(self, *, code=aiohttp.WSCloseCode.OK, message=b""):
        """
        Close all websockets
        """
        await asyncio.gather(
            *[ws.close(code=code, message=message) for ws in self.WebSockets.values()],
            return_exceptions=True
        )

    async def ping_all(self):
        """
        Ping all connected websockets
        """
        await asyncio.gather(
            *[ws.ping() for ws in self.WebSockets.values()], return_exceptions=True
        )

    async def send_str_all(self, data, compress=None):
        """
        Send string to all connected websockets
        """
        await asyncio.gather(
            *[ws.send_str(data, compress=compress) for ws in self.WebSockets.values()],
            return_exceptions=True
        )

    async def send_json_all(self, data, compress=None):
        """
        Send json to all connected websockets
        """
        await asyncio.gather(
            *[ws.send_json(data, compress=compress) for ws in self.WebSockets.values()],
            return_exceptions=True
        )

    def get(self, wsid):
        """
        Get a websocket using its wsid.

        IMPORTANT: This method can return None if the websocket has been already closed.
        """
        return self.WebSockets.get(wsid)

    async def __call__(self, request):
        ws = aiohttp.web.WebSocketResponse(
            timeout=self.Timeout,
            protocols=self.Protocols,
            compress=self.Compress,
            max_msg_size=self.MaxMsgSize,
        )

        await ws.prepare(request)

        L.log(LOG_NOTICE, "Websocket connection accepted")

        wsid = await self.register(ws, request)

        try:
            self.WebSockets[wsid] = ws
            await self.on_connect(ws, wsid)

            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.ERROR:
                    break
                else:
                    await self.on_message(ws, msg, wsid)

        except asyncio.CancelledError:
            await ws.close()

        finally:
            del self.WebSockets[wsid]
            L.log(LOG_NOTICE, "Websocket connection closed")
            await self.on_close(ws, wsid)

        return ws

    async def register(self, ws, request):
        """
        Must return the unique identifier of the websocket `wsid`.
        The default implementation is the unique counter.
        You may overload this to provide custom unique identifier
        """
        self.Counter += 1
        return self.Counter

    async def on_connect(self, websocket, wsid):
        """
        Override this method to implement action on websocket connection
        """
        pass

    async def on_message(self, websocket, message, wsid):
        """
        Override this method to receive messages from client over the websocket
        """
        pass

    async def on_close(self, websocket, wsid):
        """
        Override this method to receive a notification that client closed the websocket connection
        """
        pass



from ...abc import Expression
from ..value.valueexpr import VALUE


class FUNCTION(Expression):
    Attributes = {
        "Apply": ["*"],
    }

    Category = "Statements"

    def __init__(self, app, *, arg_apply, arg_name: str = None):
        super().__init__(app)

        if isinstance(arg_apply, Expression):
            self.Apply = arg_apply
        else:
            self.Apply = VALUE(app, value=arg_apply)

        if arg_name is not None:
            assert isinstance(arg_name, str)
            self.Name = arg_name
        else:
            self.Name = None  # Anonymous function

    def __call__(self, context, event, *args, **kwargs):
        return self.Apply(context, event, *args, **kwargs)

    def get_outlet_type(self):
        return self.Apply.get_outlet_type()



import operator

from ..abc import SequenceExpression


class ADD(SequenceExpression):
    """
    Add all values from expressions.
    """

    Attributes = {
        "Items": [
            "si64",
            "si8",
            "si16",
            "si32",
            "si128",
            "si256",
            "ui8",
            "ui16",
            "ui32",
            "ui64",
            "ui128",
            "ui256",
            "fp64",
            "fp16",
            "fp32",
            "fp128",
            "str",
        ]
    }

    Category = "Arithmetic"

    def __call__(self, context, event, *args, **kwargs):
        return self.reduce(operator.add, context, event, *args, **kwargs)

    def get_outlet_type(self):
        return _get_outlet_type_from_first(self.Items)

    def get_items_inlet_type(self):
        # TODO: This is maybe not true for integer additions
        return _get_outlet_type_from_first(self.Items)


class DIV(SequenceExpression):
    """
    Divides values in expression
    """

    Attributes = {
        "Items": [
            "si64",
            "si8",
            "si16",
            "si32",
            "si128",
            "si256",
            "ui8",
            "ui16",
            "ui32",
            "ui64",
            "ui128",
            "ui256",
            "fp64",
            "fp16",
            "fp32",
            "fp128",
        ]
    }

    Category = "Arithmetic"

    def __call__(self, context, event, *args, **kwargs):
        return self.reduce(operator.truediv, context, event, *args, **kwargs)

    def get_outlet_type(self):
        # TODO: Check if there is float among integers
        return _get_outlet_type_from_first(self.Items)

    def get_items_inlet_type(self):
        # TODO: Check if there is float among integers
        return _get_outlet_type_from_first(self.Items)


class MUL(SequenceExpression):
    """
    Multiplies values in expression.
    """

    Attributes = {
        "Items": [
            "si64",
            "si8",
            "si16",
            "si32",
            "si128",
            "si256",
            "ui8",
            "ui16",
            "ui32",
            "ui64",
            "ui128",
            "ui256",
            "fp64",
            "fp16",
            "fp32",
            "fp128",
        ]
    }

    Category = "Arithmetic"

    def __call__(self, context, event, *args, **kwargs):
        return self.reduce(operator.mul, context, event, *args, **kwargs)

    def get_outlet_type(self):
        # TODO: Check if there is float among integers
        return _get_outlet_type_from_first(self.Items)

    def get_items_inlet_type(self):
        # TODO: Check if there is float among integers
        return _get_outlet_type_from_first(self.Items)


class SUB(SequenceExpression):
    """
    Subtracts values in expression
    """

    Attributes = {
        "Items": [
            "si64",
            "si8",
            "si16",
            "si32",
            "si128",
            "si256",
            "ui8",
            "ui16",
            "ui32",
            "ui64",
            "ui128",
            "ui256",
            "fp64",
            "fp16",
            "fp32",
            "fp128",
        ]
    }

    Category = "Arithmetic"

    def __call__(self, context, event, *args, **kwargs):
        return self.reduce(operator.sub, context, event, *args, **kwargs)

    def get_outlet_type(self):
        # TODO: Check if there is float among integers
        return _get_outlet_type_from_first(self.Items)

    def get_items_inlet_type(self):
        # TODO: Check if there is float among integers
        return _get_outlet_type_from_first(self.Items)


class MOD(SequenceExpression):
    """
    Modules values in expression.
    """

    Attributes = {
        "Items": [
            "si64",
            "si8",
            "si16",
            "si32",
            "si64",
            "si128",
            "si256",
            "ui8",
            "ui16",
            "ui32",
            "ui64",
            "ui128",
            "ui256",
        ]
    }

    Category = "Arithmetic"

    def __call__(self, context, event, *args, **kwargs):
        return self.reduce(operator.mod, context, event, *args, **kwargs)

    def get_outlet_type(self):
        return _get_outlet_type_from_first(self.Items)

    def get_items_inlet_type(self):
        return _get_outlet_type_from_first(self.Items)


class POW(SequenceExpression):
    Attributes = {"Items": ["si64"]}

    Category = "Arithmetic"

    def __call__(self, context, event, *args, **kwargs):
        return self.reduce(operator.pow, context, event, *args, **kwargs)


def _get_outlet_type_from_first(items):
    if len(items) == 0:
        return int.__name__
    # Take the type of the first item in the list
    return items[0].get_outlet_type()



import pika

from bspump.asab import subscribe

from ..abc.sink import Sink
import logging


L = logging.getLogger(__name__)


class AMQPSink(Sink):
    ConfigDefaults = {
        "exchange": "amq.direct",
        "content_type": "text/plain",
        "routing_key": "",
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self._connection = pipeline.locate_connection(app, connection)
        self._channel = None
        self.Pipeline.throttle(self, True)
        self._exchange = self.Config["exchange"]
        self._content_type = self.Config["content_type"]
        self._routing_key = self.Config["routing_key"]

        app.PubSub.subscribe_all(self)

        self._connection.PubSub.subscribe(
            "AMQPConnection.open!", self._on_connection_open
        )
        self._connection.PubSub.subscribe(
            "AMQPConnection.close!", self._on_connection_close
        )
        if self._connection.ConnectionEvent.is_set():
            self._on_connection_open("simulated")

    @subscribe("Application.tick/10!")
    def _on_tick(self, event_name):
        # Heal the connection
        if self._channel is None and self._connection.ConnectionEvent.is_set():
            self._on_connection_open("simulated")

    def _on_connection_open(self, event_name):
        self._connection.Connection.channel(on_open_callback=self._on_channel_open)

    def _on_connection_close(self, event_name):
        self._channel = None
        self.Pipeline.throttle(self, True)

    def _on_channel_open(self, channel):
        self._channel = channel
        self.Pipeline.throttle(self, False)

    def process(self, context, event):
        if self._channel is None:
            raise RuntimeError("AMQP channel is not open")

        if self._exchange == "!context":
            exchange = context.get("amqp.exchange")
        else:
            exchange = self._exchange

        if exchange is None:
            L.warning("Exchange is not specified")

        try:
            self._channel.basic_publish(
                exchange,
                self._routing_key,
                event,
                pika.BasicProperties(content_type=self._content_type, delivery_mode=1),
            )
        except pika.exceptions.ChannelClosed:
            self._channel = None
            self.Pipeline.throttle(self, True)
            raise



import asyncio
import logging
import socket

from ..abc.source import Source
from ..abc.sink import Sink

#

L = logging.getLogger(__name__)


#


class DatagramSource(Source):
    """
    Description:

    """

    ConfigDefaults = {
        "address": "127.0.0.1 8888",  # IPv4, IPv6 or unix socket path
        "max_packet_size": 64 * 1024,
        "receiver_buffer_size": 0,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:


        """
        super().__init__(app, pipeline, id=id, config=config)
        self.Loop = app.Loop

        # Create a UDP socket
        self.Address = str(self.Config["address"])

        # Receive Buffer Size
        self.ReceiveBufferSize = int(self.Config["receiver_buffer_size"])

        addrline = self.Address.strip()
        if addrline.count(":") == 1:
            host, port = self.Address.rsplit(":", maxsplit=1)
            (family, socktype, proto, canonname, sockaddr) = socket.getaddrinfo(
                host, port
            )[0]

            self.Socket = socket.socket(family, socket.SOCK_DGRAM)
            self.Socket.setblocking(False)
            self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            if self.ReceiveBufferSize > 0:
                self.Socket.setsockopt(
                    socket.SOL_SOCKET, socket.SO_RCVBUF, self.ReceiveBufferSize
                )

            self.Socket.bind(sockaddr)

        else:
            self.Socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            self.Socket.setblocking(False)
            self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            if self.ReceiveBufferSize > 0:
                self.Socket.setsockopt(
                    socket.SOL_SOCKET, socket.SO_RCVBUF, self.ReceiveBufferSize
                )

            self.Socket.bind(self.Address)

        self.MaxPacketSize = int(self.Config["max_packet_size"])

    async def main(self):
        task = asyncio.ensure_future(self._receive())

        await self.stopped()

        task.cancel()
        await task

        self.Socket.close()

    async def _receive(self):
        while True:
            try:
                await self.Pipeline.ready()
                event, peer = await self.Loop.sock_recvfrom(
                    self.Socket, self.MaxPacketSize
                )
                await self.Pipeline.ready()
                await self.process(event, context={"datagram": peer})

            except asyncio.CancelledError:
                break

            except Exception:
                L.exception("Error in datagram source.")
                raise


class DatagramSink(Sink):
    """
    Description:


    """

    ConfigDefaults = {
        "address": "127.0.0.1 8888",  # IPv4, IPv6 or unix socket path
        "max_packet_size": 64 * 1024,
        "receiver_buffer_size": 0,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:


        """

        super().__init__(app, pipeline, id=id, config=config)
        self.Loop = app.Loop

        # Create a UDP socket
        self.Address = str(self.Config["address"])

        # Receive Buffer Size
        self.ReceiveBufferSize = int(self.Config["receiver_buffer_size"])

        addrline = self.Address.strip()
        if " " in addrline:
            host, port = self.Address.rsplit(" ", maxsplit=1)
        elif addrline.count(":") == 1:
            host, port = self.Address.rsplit(":", maxsplit=1)
            (family, socktype, proto, canonname, sockaddr) = socket.getaddrinfo(
                host, port
            )[0]

            self.Socket = socket.socket(family, socket.SOCK_DGRAM)
            self.Socket.setblocking(False)
            self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            if self.ReceiveBufferSize > 0:
                self.Socket.setsockopt(
                    socket.SOL_SOCKET, socket.SO_RCVBUF, self.ReceiveBufferSize
                )

            self.Socket.connect(sockaddr)
        else:
            self.Socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            self.Socket.setblocking(False)
            self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            if self.ReceiveBufferSize > 0:
                self.Socket.setsockopt(
                    socket.SOL_SOCKET, socket.SO_RCVBUF, self.ReceiveBufferSize
                )

            self.Socket.connect(self.Address)

        self.MaxPacketSize = int(self.Config["max_packet_size"])

    def process(self, context, event):
        """
        Description:


        """
        self.Socket.send(event)



import logging

import aiomysql
import pymysql
import pymysql.err

from ..abc.lookup import MappingLookup
from ..abc.lookup import AsyncLookupMixin
from ..cache import CacheDict

##

L = logging.getLogger(__name__)

##


class MySQLLookup(MappingLookup, AsyncLookupMixin):
    """
    MySQLLookup is linked with a MySQL.
    MySQLLookup provides a mapping (dictionary-like) interface to pipelines.
    MySQLLookup expects user to obtain values asynchronously in an enricher based on Generator.
    MySQLLookup feeds lookup data from MySQL database using a query.
    MySQLLookup also has a simple cache to reduce a number of database hits.

    MySQLLookup allows to specify custom cache strategy via `cache` parameter, as shown in the example below.
    LRUCacheDict removes last used elements, if the time they were lastly used exceeds the specified `max_duration` or the cache dictionary exceeds `max_size`.
    For more information, please see: https://en.wikipedia.org/wiki/Cache_replacement_policies#Least_recently_used_(LRU)

    First, it is needed to create MySQLLookup instance and register it inside the BSPump service:

            self.MySQLLookup =  MySQLLookup(self,
                    connection=mysql_connection,
                    id="MySQLLookup",
                    cache=bspump.cache.LRUCacheDict(app, max_size=1000, max_duration=1000)
                    config={
                            'from': 'user_loc',
                            'key': 'user'
                    })
            svc = app.get_service("bspump.PumpService")
            svc.add_lookup(self.MySQLLookup)

    The configuration option "from" can include a table name ...

            from="Orders"

    ...or a query string including joins like:

            from="Orders INNER JOIN Customers ON Orders.CustomerID=Customers.CustomerID"

    The MySQLLookup can be then located and used inside a custom enricher:

            class AsyncEnricher(bspump.Generator):

                    def __init__(self, app, pipeline, id=None, config=None):
                            super().__init__(app, pipeline, id, config)
                            svc = app.get_service("bspump.PumpService")
                            self.Lookup = svc.locate_lookup("MySQLLookup")

                    async def generate(self, context, event, depth):
                            if 'user' not in event:
                                    return None

                            info = await self.Lookup.get(event['user'])

                            # Inject a new event into a next depth of the pipeline
                            await self.Pipeline.inject(context, event, depth)

    """

    ConfigDefaults = {
        "statement": "*",  # Specify the statement what to select
        "from": "",  # Specify the FROM object, which can be a table or a query string
        "key": "",  # Specify key name used for search
        "query_find_one": "SELECT {} FROM {} WHERE {}=%s;",  # Specify query string to find one record in database using key
        "query_count": "SELECT COUNT(*) as 'count' FROM {};",  # Specify query string to count number of records in the database
        "query_iter": "SELECT {} FROM {};",  # Specify general query string for the iterator
    }

    def __init__(self, app, connection, id=None, config=None, cache=None):
        super().__init__(app, id=id, config=config)
        self.Connection = connection

        self.Statement = self.Config["statement"]
        self.From = self.Config["from"]
        self.Key = self.Config["key"]

        self.QueryFindOne = self.Config["query_find_one"]
        self.QueryCount = self.Config["query_count"]
        self.QueryIter = self.Config["query_iter"]

        self.Count = -1
        if cache is None:
            self.Cache = CacheDict()
        else:
            self.Cache = cache

        metrics_service = app.get_service("asab.MetricsService")
        self.CacheCounter = metrics_service.create_counter(
            "mysql.lookup.cache", tags={}, init_values={"hit": 0, "miss": 0}
        )
        self.SuccessCounter = metrics_service.create_counter(
            "mysql.lookup.success", tags={}, init_values={"hit": 0, "miss": 0}
        )

    async def _find_one(self, key):
        query = self.QueryFindOne.format(self.Statement, self.From, self.Key)
        async with self.Connection.acquire_connection() as connection:
            async with connection.cursor(aiomysql.cursors.DictCursor) as cursor_async:
                try:
                    await cursor_async.execute(query, key)
                    result = await cursor_async.fetchone()
                    return result
                except (
                    pymysql.err.InternalError,
                    pymysql.err.ProgrammingError,
                    pymysql.err.OperationalError,
                ) as e:
                    if e.args[0] in self.Connection.RetryErrors:
                        L.warning(
                            "Recoverable error '{}' occurred in MySQLLookup. Skipping lookup.".format(
                                e.args[0]
                            )
                        )
                        return None
                    raise e

    async def _count(self):
        query = self.QueryCount.format(self.From)
        async with self.Connection.acquire_connection() as connection:
            async with connection.cursor(aiomysql.cursors.DictCursor) as cursor:
                await cursor.execute(query)
                result = await cursor.fetchone()
                return result["count"]

    async def get(self, key):
        """
        Obtain the value from lookup asynchronously.
        """

        try:
            value = self.Cache[key]
            self.CacheCounter.add("hit", 1)
        except KeyError:
            value = await self._find_one(key)
            self.Cache[key] = value
            self.CacheCounter.add("miss", 1)

        if value is None:
            self.SuccessCounter.add("miss", 1)
        else:
            self.SuccessCounter.add("hit", 1)

        return value

    async def load(self):
        await self.Connection.ConnectionEvent.wait()
        self.Count = await self._count()

    def __len__(self):
        return self.Count

    def __getitem__(self, key):
        # To avoid synchronous operations completely
        raise NotImplementedError()

    def __iter__(self):
        query = self.QueryIter.format(self.Statement, self.From)
        cursor_sync = self.Connection.acquire_sync_cursor()
        cursor_sync.execute(query)
        result = cursor_sync.fetchall()
        self.Iterator = result.__iter__()
        return self

    def __next__(self):
        element = next(self.Iterator)
        key = element.get(self.Key)
        if key is not None:
            self.Cache[key] = element
        return key



import asyncio
import logging

import confluent_kafka

from ..abc.source import Source

#

L = logging.getLogger(__name__)

#


class KafkaSource(Source):
    """
            KafkaSource object consumes messages from an Apache Kafka system, which is configured in the KafkaConnection object.
            It then passes them to other processors in the pipeline.

    .. code:: python

            class KafkaPipeline(bspump.Pipeline):

                            def __init__(self, app, pipeline_id):
                                            super().__init__(app, pipeline_id)
                                            self.build(
                                                            bspump.kafka.KafkaSource(app, self, "KafkaConnection", config={'topic': 'messages'}),
                                                            bspump.kafka.KafkaSink(app, self, "KafkaConnection", config={'topic': 'messages2'}),
                                            )

            To ensure that after restart, pump will continue receiving messages where it left of, group_id has to
            be provided in the configuration.

            When the group_id is set, the consumer group is created and the Kafka server will then operate
            in the producer-consumer mode. It means that every consumer with the same group_id will be assigned
            unique set of partitions, hence all messages will be divided among them and thus unique.

            Long-running synchronous operations should be avoided or places inside the OOBGenerator in the asynchronous
            way or on thread using ASAB Proactor service (see bspump-oob-proactor.py example in "examples" folder).
            Otherwise, the session_timeout_ms should be raised to prevent Kafka from disconnecting the consumer
            from the partition, thus causing rebalance.

            Standard Kafka configuration options can be used,
            as specified in librdkafka library,
            where the options are simply passed to:

            https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md
    """

    ConfigDefaults = {
        "topic": "unconfigured",
        "refresh_topics": 0,
        "enable.auto.commit": "true",
        "auto.commit.interval.ms": "1000",
        "auto.offset.reset": "smallest",
        "group.id": "bspump",
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        """
        Initializes parameters.

        **Parameters**

        app : Application
                        Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html#>`_.

        pipeline : Pipeline
                        Name of the Pipeline.

        connection : Connection
                        information needed to create a connection.

        id : , default = None

        config : , default = None

        """
        super().__init__(app, pipeline, id=id, config=config)

        self.App = app
        self.Connection = self.Pipeline.locate_connection(app, connection)
        self.Sleep = 100 / 1000.0
        self.ConsumerConfig = {}

        self.SpecialKeys = frozenset(["oauth_cb"])

        # Copy connection options
        for key, value in self.Connection.Config.items():
            if key in self.SpecialKeys:
                self.ConsumerConfig[key] = value

            else:
                self.ConsumerConfig[key.replace("_", ".")] = value

        # Copy configuration options, avoid the topic
        for key, value in self.Config.items():
            if key == "topic" or key == "refresh_topics":
                continue

            if key in self.SpecialKeys:
                self.ConsumerConfig[key] = value

            else:
                self.ConsumerConfig[key.replace("_", ".")] = value

        # Create subscription list
        self.Subscribe = []

        for s in self.Config["topic"].split(" "):
            if ":" not in s:
                self.Subscribe.append(s)
            else:
                self.Subscribe.append(s.rsplit(":", 1))

        self.Running = True

        # For refreshing of topics/subscription
        self.RefreshTopics = int(self.Config["refresh_topics"])
        self.LastRefreshTopicsTime = self.App.time()

    async def main(self):
        while self.Running:
            try:
                c = confluent_kafka.Consumer(self.ConsumerConfig, logger=L)

            except BaseException as e:
                L.exception("Error when connecting to Kafka")
                self.Pipeline.set_error(None, None, e)
                return

            c.subscribe(self.Subscribe)

            try:
                while 1:
                    await self.Pipeline.ready()
                    current_time = self.App.time()

                    if (
                        self.RefreshTopics > 0
                        and current_time
                        > self.LastRefreshTopicsTime + self.RefreshTopics
                    ):
                        L.info("Topics refreshed in '{}'.".format(self.Id))
                        c.unsubscribe()
                        c.close()
                        self.LastRefreshTopicsTime = current_time
                        break

                    m = c.poll(0.2)

                    if m is None:
                        await asyncio.sleep(self.Sleep)
                        continue

                    if m.error():
                        L.error(
                            "The following error occured while polling for messages: '{}'.".format(
                                m.error()
                            )
                        )
                        await asyncio.sleep(self.Sleep)
                        continue

                    await self.process(
                        m.value(),
                        context={
                            "kafka_key": m.key(),
                            "kafka_headers": m.headers(),
                            "_kafka_topic": m.topic(),
                            "_kafka_partition": m.partition(),
                            "_kafka_offset": m.offset(),
                        },
                    )

            except asyncio.CancelledError:
                self.Running = False

            except BaseException as e:
                L.exception("Error when processing Kafka message")
                self.Pipeline.set_error(None, None, e)



from .sink import SlackAbcSink, SlackMessageSink, SlackTextSink
from .connection import SlackConnection


__all__ = (
    "SlackAbcSink",
    "SlackMessageSink",
    "SlackTextSink",
    "SlackConnection",
)



from ..abc import SequenceExpression, Expression
from ..declerror import DeclarationError
from .value.valueexpr import VALUE


class AND(SequenceExpression):
    """
    Checks if all expressions are true, respectivelly, stop on the first False
    """

    Attributes = {"Items": ["bool"]}

    Category = "Logic"

    def __call__(self, context, event, *args, **kwargs):
        for item in self.Items:
            try:
                v = item(context, event, *args, **kwargs)
            except Exception as e:
                raise DeclarationError(
                    original_exception=e, location=item.get_location()
                )
            if v is None or not v:
                return False
        return True

    def get_outlet_type(self):
        return bool.__name__

    def get_items_inlet_type(self):
        return bool.__name__


class OR(SequenceExpression):
    """
    Checks if at least one of the expressions is true
    """

    Attributes = {"Items": ["bool"]}

    Category = "Logic"

    def __call__(self, context, event, *args, **kwargs):
        for item in self.Items:
            try:
                v = item(context, event, *args, **kwargs)
            except Exception as e:
                raise DeclarationError(
                    original_exception=e, location=item.get_location()
                )
            if v is not None and v:
                return True
        return False

    def get_outlet_type(self):
        return bool.__name__

    def get_items_inlet_type(self):
        return bool.__name__


class NOT(Expression):
    """
    Returns inverse value of the expression
    """

    Attributes = {
        "What": "bool",
    }

    Category = "Logic"

    def __init__(self, app, *, arg_what):
        super().__init__(app)
        self.What = arg_what

        if isinstance(arg_what, Expression):
            self.What = arg_what
        else:
            self.What = VALUE(app, value=arg_what)

    def __call__(self, context, event, *args, **kwargs):
        try:
            return not self.What(context, event, *args, **kwargs)
        except TypeError:
            return False

    def get_outlet_type(self):
        return bool.__name__



import re
import logging
import functools

import aiohttp.web

#

L = logging.getLogger(__name__)

#


def required(*resources):
    """
    OBSOLETE

    Checks that user authorized with access token in
    Authorization header has access to a given tenant space
    using SeaCat Auth RBAC authorization.

    It uses a cache to limit the number of HTTP checks.

    Example of use:

    @asab.web.tenant.tenant_handler
    @asab.web.authz.required("tenant:access")
    async def endpoint(self, request, *, tenant):
            ...
    """

    def decorator_required(func):
        @functools.wraps(func)
        async def wrapper(*args, **kargs):
            request = args[-1]

            # Obtain authz service from the request
            authz_service = request.AuthzService

            if not request.AuthzService.is_ready():
                L.error("Cannot authorize request - AuthzService is not ready.")
                raise aiohttp.web.HTTPUnauthorized()

            bearer_token = _get_bearer_token(request)

            # For resistancy against security attacks
            if bearer_token is None:
                raise aiohttp.web.HTTPUnauthorized()

            if await authz_service.authorize(
                resources=resources,
                bearer_token=bearer_token,
                tenant=getattr(request, "Tenant", None),
            ):
                return await func(*args, **kargs)

            # Be defensive
            raise aiohttp.web.HTTPUnauthorized()

        return wrapper

    return decorator_required


def userinfo_handler(func):
    """
    Fetches userinfo and passes the response dict to the decorated function.

    It uses a cache to limit the number of HTTP checks.

    Example of use:

    @asab.web.tenant.tenant_handler
    @asab.web.authz.userinfo
    async def endpoint(self, request, *, tenant, userinfo):
            ...
    """

    @functools.wraps(func)
    async def wrapper(*args, **kargs):
        request = args[-1]

        # Obtain authz service from the request
        authz_service = request.AuthzService

        bearer_token = _get_bearer_token(request)

        # Fail if no access token is found in the request
        if bearer_token is None:
            L.warning(
                "Access token has not been provided in the request - unauthorized."
            )
            raise aiohttp.web.HTTPUnauthorized()

        userinfo_data = authz_service.userinfo(bearer_token=bearer_token)
        if userinfo_data is not None:
            return await func(*args, userinfo=userinfo_data, **kargs)

        # Be defensive
        L.warning("Failure to get userinfo  - unauthorized.")
        raise aiohttp.web.HTTPUnauthorized()

    return wrapper


def _get_bearer_token(request):
    authorization_header_rg = re.compile(r"^\s*Bearer ([A-Za-z0-9\-\.\+_~/=]*)")

    authorization_value = request.headers.get(aiohttp.hdrs.AUTHORIZATION, None)
    bearer_token = None

    # Obtain access token from the authorization header
    if authorization_value is not None:
        authorization_match = authorization_header_rg.match(authorization_value)
        if authorization_match is not None:
            bearer_token = authorization_match.group(1)

    return bearer_token
